"""
Parser classes for the Australia's beamlines.

Current facilities supported include are exclusively:
- Australian Synchrotron.
"""

from pyNexafs.parsers.au.aus_sync import (
    SXR_NEXAFS,
    MEX1_NEXAFS,
    MEX1_to_QANT_AUMainAsc,
    MEX2_NEXAFS,
    MEX2_to_QANT_AUMainAsc,
)

__all__ = [
    "SXR_NEXAFS",
    "MEX1_NEXAFS",
    "MEX1_to_QANT_AUMainAsc",
    "MEX2_NEXAFS",
    "MEX2_to_QANT_AUMainAsc",
]



"""
Parser classes for the Soft X-ray (SXR) beamline at the Australian Synchrotron.
"""

from pyNexafs.parsers import parserBase
from pyNexafs.types import parse_fn_ret_type
from pyNexafs.utils.mda import MDAFileReader
from typing import Any
import typing
from numpy.typing import NDArray
import io
import numpy as np
from typing import override


class SXR_NEXAFS(parserBase):
    """
    Australian Synchrotron Soft X-ray (SXR) NEXAFS parser.

    Parses data formats including '.asc' and '.mda' formats from the SXR
    Near Edge X-ray Absorption Fine Structure (NEXAFS) tool.

    Attributes
    ----------
    ALLOWED_EXTENSIONS
    COLUMN_ASSIGNMENTS
    SUMMARY_PARAM_RAW_NAMES
    RELABELS

    Notes
    -----
    Implemented for data as of 2024-Mar.
    """

    ALLOWED_EXTENSIONS = [
        ".asc",  # Ascii files exported from mda2asc tool
        ".mda",  # Binary MDA files from EPICS/Motor scans
        ".txt",  # Files generated by QANT software
    ]  # ascii files exported from the binary .mda files generated.

    COLUMN_ASSIGNMENTS = {
        "x": "SR14ID01PGM_CALC_ENERGY_MONITOR.P",
        "y": [
            # "SR14ID01PGM:LOCAL_SP",
            "SR14ID01IOC68:scaler1.S20",  # Drain Current VF
            "SR14ID01IOC68:scaler1.S17",  # Direct PHD VF
            "SR14ID01IOC68:scaler1.S18",  # I0 VF
            "SR14ID01IOC68:scaler1.S19",  # Ref Foil VF
            "SR14ID01IOC68:scaler1.S21",  # Channeltron Front (PEY)
            "SR14ID01IOC68:scaler1.S22",  # MCP (TFY)
            "SR14ID01IOC68:scaler1.S23",  # Hemispherical Analyser (AEY)
            # No longer included in files...?
            # "SR14ID01IOC68:scaler1.S3", # I0 VF
            # "SR14ID01IOC68:scaler1.S4", # Ref Foil VF
            # "SR14ID01IOC68:scaler1.S8", # Direct PHD VF
        ],
        "y_errs": None,
        "x_errs": None,
    }

    SUMMARY_PARAM_RAW_NAMES = [
        "SR14ID01NEXSCAN:saveData_comment1",
        "SR14ID01NEXSCAN:saveData_comment2",
        "SR14ID01NEX01:R_MTR.RBV",
        "SR14ID01NEX01:RULER_ID",
        "SR14ID01:BRANCH_MODE",
        "SR14ID01NEX01:C_MTR.RBV",
        "SR14ID01NEX01:Z_MTR.RBV",
        "SR14ID01NEX01:X_MTR.RBV",
        "SR14ID01NEX01:Y_MTR.RBV",
    ]

    RELABELS = {
        "SR14ID01PGM:REMOTE_SP": "Photon Energy",
        "SR14ID01PGM:LOCAL_SP": "Local Energy Setpoint",
        "SR14ID01PGM_ENERGY_SP": "Energy Setpoint",
        "SR14ID01PGM_CALC_ENERGY_MONITOR.P": "Encoder Photon Energy",
        (
            "SR14ID01IOC68:scaler1.TPSR14ID01DTACQ06:sample_time",
            "SR14ID01DTACQ07:sample_time",
        ): "Exp Time",
        (
            "SR14ID01IOC68:scaler1.S2",
            "SR14ID01IOC68:scaler1.S20",
            "SR14ID01DTACQ06:ch4_v2f_ave_rbv",
            "SR14ID01DTACQ07:ch3_v2f_ave_rbv",
            "Drain Current VF",
        ): "Drain Current",  # Previously labelled with a VF
        (
            "SR14ID01IOC68:scaler1.S3",
            "SR14ID01IOC68:scaler1.S18",
            "SR14ID01DTACQ06:ch3_v2f_ave_rbv",  # Izero-VF
            "SR14ID01DTACQ07:ch2_v2f_ave_rbv",
            "I0 VF",
        ): "I0",
        (
            "SR14ID01IOC68:scaler1.S4",
            "SR14ID01IOC68:scaler1.S19",
            "SR14ID01DTACQ06:ch2_v2f_ave_rbv",
            "SR14ID01DTACQ07:ch1_v2f_ave_rbv",
            "Ref Foil VF",
        ): "Ref Foil",
        (
            "SR14ID01IOC68:scaler1.S6",
            "SR14ID01IOC68:scaler1.S22",
            "SR14ID01DTACQ06:ch1_counts_rbv",
            "SR14ID01DTACQ07:ch2_counts_rbv",
            "MCP (TFY)",
        ): "Micro Channel Plate (TFY)",
        (
            "SR14ID01IOC68:scaler1.S8",
            "SR14ID01IOC68:scaler1.S17",
            "SR14ID01DTACQ06:ch5_v2f_ave_rbv",
            "SR14ID01DTACQ07:ch4_v2f_ave_rbv",
            "Direct PHD VF",
        ): "Direct PHD",
        (
            "SR14ID01IOC68:scaler1.S9",
            "SR14ID01DTACQ06:ch1_v2f_ave_rbv",
            "BL PHD VF",
        ): "BL PhD",
        (
            "SR14ID01IOC68:scaler1.S10",
            "SR14ID01DTACQ07:ch1_counts_rbv",
        ): "Channeltron Front (PEY)",
        (
            "SR14ID01IOC68:scaler1.S21",
            "SR14ID01DTACQ06:ch2_counts_rbv",
        ): "Channeltron",
        (
            "SR14ID01IOC68:scaler1.S11",
            "TFT PHD VF",
        ): "TFY PHD",
        (
            "SR14ID01IOC68:scaler1.S23",
            "SR14ID01DTACQ07:ch3_counts_rbv",
            "Hemispherical Analyser AEY",
        ): "Hemispherical Analyser (AEY)",
        "SR14ID01AMP01:CURR_MONITOR": "Drain Current - Keithley1",
        "SR14ID01AMP02:CURR_MONITOR": "BL PHD - Keithley2",
        "SR14ID01AMP03:CURR_MONITOR": "I0 - Keithley3",
        "SR14ID01AMP04:CURR_MONITOR": "Ref Foil -Keithley4",
        "SR14ID01AMP05:CURR_MONITOR": "Direct PHD - Keithley5",
        "SR14ID01AMP06:CURR_MONITOR": "Keithley6",
        "SR14ID01AMP07:CURR_MONITOR": "Ref Foil - Keithley7",
        "SR14ID01AMP08:CURR_MONITOR": "Drain Current - Keithley8",
        "SR14ID01AMP09:CURR_MONITOR": "I0 - Keithley9",
        "SR14ID01:BL_GAP_REQUEST": "Undulator Gap Request",
        "SR14ID01:GAP_MONITOR": "Undulator Gap Readback",
        "SR11BCM01:CURRENT_MONITOR": "Ring Current",
        "SR14ID01NEXSCAN:saveData_comment1": "Note 1",
        "SR14ID01NEXSCAN:saveData_comment2": "Note 2",
        "SR14ID01:BRANCH_MODE": "Branch",
        "SR14ID01NEX01:RULER_ID": "Ruler ID",
        "SR14ID01NEX01:R_MTR.RBV": "R",
        "SR14ID01NEX01:C_MTR.RBV": "C",
        "SR14ID01NEX01:Z_MTR.RBV": "Z",
        "SR14ID01NEX01:X_MTR.RBV": "X",
        "SR14ID01NEX01:Y_MTR.RBV": "Y",
        "SR14ID01DTACQ06:ch1_converted_ave_rbv": "BL PHD - SR570-2",
        "SR14ID01DTACQ06:ch2_converted_ave_rbv": "Ref Foil - SR570-4",
        "SR14ID01DTACQ06:ch3_converted_ave_rbv": "I0 - SR570-3",  # Izero-SR570-3"
        "SR14ID01DTACQ06:ch4_converted_ave_rbv": "Drain Current SR570-1",
        "SR14ID01DTACQ06:ch5_converted_ave_rbv": "Direct PHD - SR570-5",
        "SR14ID01DTACQ07:ch1_converted_ave_rbv": "Ref Foil - SR570-8",
        "SR14ID01DTACQ07:ch2_converted_ave_rbv": "I0 - SR570-7",
        "SR14ID01DTACQ07:ch3_converted_ave_rbv": "Drain_Current - SR570-6",
        "SR14ID01DTACQ07:ch4_converted_ave_rbv": "Drain Current - SR570-9",
        "SR14ID01DPP01:mca1.R0": "Dante ROI (PFY)",
        "SR14ID01DPP01:ch1:InputCountRate": "Dante Input Count Rate (TFY)",
        "SR14ID01DPP01:ch1:OutputCountRate": "Dante Output Count Rate (TFY)",
    }

    @classmethod
    def parse_asc_202403(
        cls, file: typing.TextIO, header_only: bool = False
    ) -> parse_fn_ret_type:
        """
        Read Australian Synchrotron SXR NEXAFS `.asc` files.

        Parameters
        ----------
        file : typing.TextIO
            TextIO of the datafile (i.e. open('file.asc', 'r')).
        header_only : bool, optional
            If True, then only the header of the file is read and NDArray is returned as None, by default False.

        Returns
        -------
        tuple[NDArray, list[str], dict[str, Any]]
            Returns a set of data as a numpy array,
            labels as a list of strings,
            units as a list of strings,
            and parameters as a dictionary.

        Raises
        ------
        ValueError
            If the file is not a valid .asc file.
        """

        # Initialise structures
        params = {"filename": file.name}
        labels = []
        units = []

        # Check valid format.
        if not file.name.endswith(".asc"):
            raise ValueError(f"File {file.name} is not a valid .asc file.")

        ### Read file
        # Check header is correct
        assert file.readline() == "## mda2ascii 0.3.2 generated output\n"

        # skip 2 lines
        [file.readline() for i in range(2)]

        ## 1 Initial Parameters including MDA File v, Scan Num, Ovrl scn dim, scn size
        for i in range(4):
            line = file.readline()[2:].strip().split(" = ", 1)
            if len(line) == 2:
                params[line[0]] = line[1]

        # skip 2 lines
        [file.readline() for i in range(2)]
        # Check PV header is correct
        assert file.readline() == "#  Extra PV: name, descr, values (, unit)\n"

        # skip 1 line
        file.readline()

        ## 2 Extra Parameter/Values
        line = file.readline()
        while line != "\n":
            # Discard first 11 characters and linenumber '# Extra PV 1:'.
            line = line.strip().split(":", 1)[1]
            # Separate the name, description, values and units.
            line = line.split(",")
            # Trim values and append to params in a tuple.
            vals = [line[i].strip().replace('"', "") for i in range(1, len(line))]
            # Convert string values to float / int if possible:
            if len(vals) == 4:  # has units column
                try:
                    vals[3] = float(vals[3]) if "." in vals[3] else int(vals[3])
                except ValueError:
                    pass
            # Add vals to params
            params[line[0].strip()] = vals

            # read the newline
            line = file.readline()

        # skip 1 extra line
        file.readline()

        ## 3 Scan Header:
        assert file.readline() == "# 1-D Scan\n"
        for i in range(3):
            line = file.readline()[2:].strip().split(" = ", 1)
            if len(line) == 2:
                params[line[0]] = line[1]

        # skip 1 line
        file.readline()

        ## 4 Scan Column properties:
        # Labels for columns describing data columns.
        # i.e.
        # i.e. #  Positioner: name, descr, step mode, unit, rdbk name, rdbk descr, rdbk unit
        column_types = {}
        line = file.readline()
        while line != "\n":
            line = line.strip().split(": ", 1)
            col_name = line[0][3:]  # ignore "#  " before name
            col_attrs = line[1].split(", ")
            column_types[col_name] = col_attrs
            # read the newline
            line = file.readline()

        # skip 1 line
        file.readline()

        # Column descriptions
        # i.e. #    2  [1-D Positioner 1]  SR14ID01PGM:LOCAL_SP, Mono setpoint, TABLE, eV, SR14ID01PGM:LOCAL_SP, Mono setpoint, eV
        column_descriptions = {}
        column_type_assignments = {}
        line = file.readline()
        while line != "\n":
            # Take index info
            index = int(line[1:6].strip())
            index_line = (
                index == 1
            )  # Boolean to determine if on index description line.
            # Take coltype info
            desc_type = line[8:26]
            assert (
                desc_type[0] == "[" and desc_type[-1] == "]"
            )  # Check parenthesis of the line parameters
            desc_type = desc_type[1:-1].strip()
            # Check if valid col type
            valid = False if not index_line else True
            for coltype in column_types:
                if coltype in desc_type:
                    valid = True
                    break
            if not valid:
                raise ValueError(f"Invalid column type {desc_type} in line {line}")

            # Take info
            desc_info = line[28:].split(", ")
            column_descriptions[index] = desc_info
            column_type_assignments[index] = desc_type
            # Check that the initial parameter begins with the Instrument descriptor.
            if not index_line:
                assert desc_info[0].startswith("SR14ID01")  # code for initial
            # Add to labels and units to lists
            labels += [desc_info[0].strip()] if not index_line else ["Index"]
            if "Positioner" in desc_type:
                pUnit = desc_info[3].strip()  # hardcoded 'unit' position
                units += [pUnit]
            elif "Detector" in desc_type:
                dUnit = desc_info[2].strip()  # hardcoded 'unit' position
                units += [dUnit if dUnit != "" else None]
            elif index_line:
                units += [None]

            # read next line
            line = file.readline()

        # add column data to params
        params["column_types"] = column_types
        params["column_descriptions"] = column_descriptions
        params["column_type_assignments"] = column_type_assignments

        if header_only:
            # Do not process remaining lines
            return None, labels, units, params

        ## 5 Scan Values
        assert file.readline() == "# 1-D Scan Values\n"
        lines = file.readlines()  # read remaining lines efficiently

        # Convert data to numpy array.
        data = np.loadtxt(lines, delimiter="\t")
        # [np.loadtxt(line, delimiter="\t")
        #     for line in lines]
        data = np.array(data)

        return data, labels, units, params

    @classmethod
    def parse_mda_2024_03(
        cls, file: typing.IO | str, header_only: bool = False
    ) -> parse_fn_ret_type:
        """
        Read Australian Synchrotron SXR NEXAFS .mda files.

        Parameters
        ----------
        file : typing.IO | str
            IO of the datafile (i.e. open('file.mda', 'rb')).
        header_only : bool, optional
            If True, only the header of the file is read and NDArray is returned as None, by default False.

        Returns
        -------
        data : NDArray | None
            The data array if header_only is False, else None.
        labels : list[str]
            List of labels for each data column.
        units : list[str]
            List of units for each data column.
        params : dict[str, Any]
            A dictionary of parameters extracted from the file, or an empty dictionary if no parameters are returned.

        Raises
        ------
        ValueError
            If the file is not a valid .mda file.
        """

        # Initialise parameter list
        params = {}
        labels = []

        if isinstance(file, str):
            # Open the file in binary mode
            with open(file, "rb") as f:
                return cls.parse_mda_2024_03(f, header_only=header_only)
        elif isinstance(file, (typing.TextIO, io.TextIOWrapper)):
            return cls.parse_mda_2024_03(file.buffer, header_only=header_only)
        else:
            assert isinstance(file, (typing.BinaryIO, io.BufferedReader, io.BytesIO)), (
                f"File must be a typing.BinaryIO implementation or file path string. Was '{type(file)}'."
            )

        # Check valid format.
        if not file.name.endswith(".mda"):
            raise ValueError(f"File {file.name} is not a valid .mda file.")

        mda = MDAFileReader(file)

        mda_header = mda.read_header_as_dict()
        if mda_header["mda_rank"] != 1:
            raise ValueError("MDA file is not 1D, incompatible for regular NEXAFS.")
        mda_params = mda.read_parameters()
        mda_arrays, mda_scans = mda.read_scans(header_only=header_only)

        # Add values to params dict
        params.update(mda_header)
        params.update(mda_params)
        # Add column types and descriptions to params.
        scan_1d = mda_scans[0]
        column_types = {
            "Positioner": (
                "name",
                "descr",
                "step mode",
                "unit",
                "rdbk name",
                "rdbk descr",
                "rdbk unit",
            ),
            "Detector": (
                "name",
                "descr",
                "unit",
            ),
        }
        column_descriptions = {
            i: [
                p.name,
                p.desc,
                p.step_mode,
                p.unit,
                p.readback_name,
                p.readback_desc,
                p.readback_unit,
            ]
            for i, p in enumerate(scan_1d.positioners)
        }
        column_descriptions.update(
            {
                i + len(scan_1d.positioners): [d.name, d.desc, d.unit]
                for i, d in enumerate(scan_1d.detectors)
            }
        )
        params["column_types"] = column_types
        params["column_descriptions"] = column_descriptions
        # Collect units and labels:
        labels = []
        units = []
        for i, p in enumerate(scan_1d.positioners):
            labels.append(p.name)
            units.append(p.unit)
        for i, d in enumerate(scan_1d.detectors):
            labels.append(d.name)
            units.append(d.unit)

        if header_only:
            return None, labels, units, params
        data: NDArray | tuple[NDArray] | None
        if isinstance(mda_arrays, list):
            if len(mda_arrays) == 1:
                data = mda_arrays[0]
            else:
                data = tuple(*mda_arrays)
        else:
            data = None
        return data, labels, units, params

    @classmethod
    def parse_txt_qant_2025_12(
        cls, file: typing.TextIO | str, header_only: bool = False
    ) -> parse_fn_ret_type:
        """
        A QANT exported txt file reader for SXR NEXAFS data.

        Works for outputs from QANTv1.14.

        Parameters
        ----------
        file : typing.TextIO
            TextIO of the datafile (i.e. open('file.txt', 'r')).
        header_only : bool, optional
            If True, only the header of the file is read and NDArray is returned as None, by default False.

        Returns
        -------
        Data : None or NDArray
            The data array if header_only is False, else None.
        labels : list[str]
            List of labels for each data column.
        units : list[str]
            List of units for each data column.
        tuple[NDArray | None, list[str], list[str], dict[str, Any]]
            A tuple containing the data array, labels as a list of strings,
            units as a list of strings,
            and parameters as a dictionary.

        Raises
        ------
        ValueError
            If the file is not a valid .txt file.
        """
        if isinstance(file, str):
            # Open the file in text mode
            with open(file, "r") as f:
                return cls.parse_txt_qant_2025_12(f, header_only=header_only)

        # Initialise structures
        params = {"filename": file.name}
        labels = []
        units = None

        # Check valid format.
        if not file.name.endswith(".txt"):
            raise ValueError(f"File {file.name} is not a valid .txt file.")

        ### Read file
        # Check header is correct
        assert file.readline().startswith("File Produced by QANT")
        separator_line = "----------------------------------------------------\n"
        assert file.readline() == separator_line
        assert file.readline() == "List of dataseries and their notes:\n"

        # skip 1 lines
        file.readline()

        # Skip noted information about the normalisation steps for each channel
        line = file.readline()
        while line != separator_line:
            if line == "":  # "" is EOF.
                return None, labels, units, params
            line = file.readline()

        # Collect the parameters.
        assert file.readline() == "List of strings or notes regarding this dataset:\n"
        line = file.readline()
        separator_line = (
            "-----------------------------------------------------\n"  # note extra -
        )
        while line != separator_line and line != "":  # "" is EOF.
            if line == "":  # "" is EOF.
                return None, labels, units, params
            # Process the line
            if line != "\n":
                if "=" in line:
                    # Split the line based on the first '=' only.
                    key, val = line.split("=", maxsplit=1)
                    params[key.strip()] = val.strip()
            line = file.readline()

        # Numerical values
        line = file.readline()
        assert line == "List of numeric variables regarding this dataset:\n", (
            f"Unexpected line: {line}"
        )
        line = file.readline()
        while line != separator_line and line != "":  # "" is EOF.
            if line == "":  # "" is EOF.
                return None, labels, units, params
            # Don't process these lines.
            line = file.readline()

        assert file.readline() == "Extra PVs:\n"
        file.readline()  # skip header line

        ## 2 Extra Parameter/Values
        data_line = "Data---------------------------------------------------\n"

        line = file.readline()
        while line != data_line and line != "":  # "" is EOF.
            if line == "":  # "" is EOF.
                return None, labels, units, params
            elif line == "\n" or "\t" not in line:
                # Skip empty lines
                line = file.readline()
                continue

            # Separate the name, description, values and units.
            line = line.split("\t")
            # Trim values and append to params in a tuple.
            vals = [line[i].strip().replace('"', "") for i in range(1, len(line))]

            # Convert string values to float / int if possible:
            if len(vals) == 4:  # has units column
                try:
                    vals[2] = float(vals[2]) if "." in vals[2] else int(vals[2])
                except ValueError:
                    pass
            # Add vals to params
            params[line[0].strip()] = vals

            # read the newline
            line = file.readline()

        # Skip two lines
        [file.readline() for i in range(2)]

        header_line = file.readline()
        assert header_line != "" and "\t" in header_line, (
            "No data header found in QANT .txt file."
        )
        labels = header_line.replace("'", "").strip().split("\t")

        # No units provided.

        if header_only:
            # Do not process remaining lines
            return None, labels, units, params

        ## 5 Scan Values
        lines = file.readlines()  # read remaining lines efficiently
        # Remove empty lines
        lines = [line for line in lines if line.strip() != ""]
        # Convert data to numpy array.
        data = np.loadtxt(lines, delimiter="\t")
        # [np.loadtxt(line, delimiter="\t")
        #     for line in lines]
        data = np.array(data)
        assert data.shape[0] != 0, "No data found in QANT .txt file."
        return data, labels, units, params

    @property
    @override
    def summary_param_values(self) -> list[Any]:
        """
        Return a list of important parameter values of the data file.

        Uses the list element corresponding to 'value' for each file.
        Overrides base method to use the 'value' element of the SXR parameter list.

        Returns
        -------
        list
            List of important parameter values.
        """
        # Get second element which is the parameter number.
        return [self.params[key][1] for key in self.SUMMARY_PARAM_RAW_NAMES]

    @property
    @override
    def summary_param_names_with_units(self) -> list[str]:
        """
        Return a list of important parameter names with units.

        Requires a loaded dataset to return the units of the parameters.
        Not a pre-defined class method.

        Returns
        -------
        list
            List of important parameter names with units.
        """
        pNames = self.SUMMARY_PARAM_RAW_NAMES
        pUnits = [
            (
                self.params[pName][2]
                if (
                    self.params is not None  # Params loaded
                    and pName in self.params  # Parameter listed
                    and hasattr(
                        self.params[pName], "__len__"
                    )  # Parameter has a list of values
                    and len(self.params[pName])
                    == 3  # 3 params for value, description, unit
                    and self.params[pName][2] != ""  # Unit value is not empty.
                )
                else None
            )
            for pName in pNames
        ]
        if not self.relabel:
            return [
                f"{pName} ({unit})" if unit is not None else pName
                for pName, unit in zip(pNames, pUnits)
            ]
        else:
            relabel_names = []
            for pName, unit in zip(pNames, pUnits):
                if unit is not None:
                    relabel_names.append(
                        f"{self.RELABELS[pName]} ({unit})"
                        if pName in self.RELABELS
                        else f"{pName} ({unit})"
                    )
                else:
                    relabel_names.append(
                        self.RELABELS[pName] if pName in self.RELABELS else pName
                    )
            return relabel_names


if __name__ == "__main__":
    # Test each of the parser functions with the test data.
    fname = "sxr129582-QANT_raw.txt"
    fname = "sxr130214-QANT_edge_norm.txt"
    fname = "sxr130214.asc"
    fname = "sxr130214.mda"

    #
    import pkgutil
    import io

    bdata = pkgutil.get_data("pyNexafs", f"../tests/test_data/au/SXR/2024-03/{fname}")
    assert bdata is not None
    bio = io.BytesIO(bdata)
    fdata = io.TextIOWrapper(bio, encoding="utf-8")
    result = SXR_NEXAFS.parse_txt_qant_2025_12(fdata, header_only=False)
    print(result)



"""
Module to perform graphical NEXAFS fitting
"""

from PyQt6.QtWidgets import QWidget, QVBoxLayout, QLabel
from pyNexafs.gui.widgets.graphing.matplotlib.graphs import FigureCanvas, NEXAFS_NavQT
from pyNexafs.nexafs import scanBase
from matplotlib.figure import Figure

# Extras for __main__ testing
import sys
import os
from PyQt6.QtWidgets import QApplication
from pyNexafs.parsers.au.aus_sync.MEX2 import MEX2_NEXAFS


class globalFitter(QWidget):
    """
    Global fitting class for a set of NEXAFS spectrums.
    """

    def __init__(self, parent: QWidget | None = None, scans: list[scanBase] = []):
        # Initialise widget
        super().__init__(parent=parent)

        # Setup layout
        layout = QVBoxLayout()
        self.setLayout(layout)

        # Setup margins if has a parent.
        if parent is not None:
            layout.setContentsMargins(0, 0, 0, 0)

        # Save the scan list
        self._scans = scans.copy()

        # Setup the UI elements
        fig = Figure()
        fig_canvas = FigureCanvas(fig)
        fig_canvas_nav = NEXAFS_NavQT(canvas=fig_canvas)

        # Add elements to the UI
        layout.addWidget(QLabel("GLobal Fitting"))
        layout.addWidget(fig_canvas)
        layout.addWidget(fig_canvas_nav)


if __name__ == "__main__":
    # Load some example parsers
    base_dir = os.getcwd()
    f1 = r"tests\\test_data\\au\\MEX2\\MEX2_5640.mda"
    f2 = r"tests\\test_data\\au\\MEX2\\MEX2_5642.mda"
    files = [os.path.normpath(os.path.join(base_dir, f)) for f in [f1, f2]]
    energy_bins = (2135, 2560)  # eV
    parsers = [
        MEX2_NEXAFS(f, header_only=False, energy_bin_domain=energy_bins) for f in files
    ]
    scans = [p.to_scan() for p in parsers]

    app = QApplication(sys.argv)
    window = globalFitter(parent=None, scans=scans)
    window.show()
    window.setWindowTitle("Summary Param Selector")
    sys.exit(app.exec())



from PyQt6.QtWidgets import (
    QWidget,
    QVBoxLayout,
    QLabel,
    QListWidget,
    QSplitter,
)
from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QApplication
import sys
import matplotlib
import matplotlib.axes
import matplotlib.figure

from pyNexafs.nexafs.scan import scanAbstract
from pyNexafs.parsers import parserBase
from pyNexafs.gui.widgets.graphing.matplotlib.graphs import FigureCanvas, NavTBQT

# from pyNexafs.gui.widgets.normaliser import NavTBQT_Norm, normaliserSettings

import numpy as np
from typing import Type


class nexafsViewer(QWidget):
    """
    General viewer for NEXAFS parser objects.

    Takes
    Includes the use of a `normalisingGraph` to display and normalise NEXAFS data.


    """

    def __init__(self, parent=None):  # norm_settings: normaliserSettings = None
        super().__init__(parent)
        # self._norm_settings = norm_settings

        # Initialise elements
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)
        if parent is not None:
            self.setContentsMargins(0, 0, 0, 0)
            self._layout.setContentsMargins(0, 0, 0, 0)
        self._draggable = QSplitter(Qt.Orientation.Vertical)

        # Initialise properties
        self._selected_filenames = []
        self._scan_objects = {}
        self._dataseries_list = []  # List of common labels in current selection.

        # Initialise viewing elements
        self._dataseries_list_view = QListWidget()
        self._normGraph = normalisingGraph()

        # Add to layout
        self._draggable.addWidget(QLabel("Data Series"))
        self._draggable.addWidget(self._dataseries_list_view)
        self._draggable.addWidget(self._normGraph)
        self._layout.addWidget(self._draggable)

        # Attributes
        self._dataseries_list_view.setSelectionMode(
            QListWidget.SelectionMode.ExtendedSelection
        )

        # Connections
        self._dataseries_list_view.itemSelectionChanged.connect(
            self.on_label_selection_change
        )

        # Boolean flag to prevent selection change from triggering callback.
        self._updating = False

    @property
    def dataseries(self) -> list[str]:
        """
        Returns the y_label intersection of currently selected scan objects.

        Returns
        -------
        list[str]
            Labels common to selected scan objects.
        """
        return self._dataseries_list

    def add_parsers_to_scans(
        self,
        parsers: dict[str, type[parserBase]] | list[type[parserBase]],
        load_all_columns: bool = True,
        override_scans: bool = False,
    ) -> None:
        """
        Add parsers to the scan list, through conversion to scan objects.

        If an scan object exists with the same filename, the `override_scans` flag will determine whether to replace the existing reference.

        Parameters
        ----------
        parsers : dict[str, scan_abstract] | list[type[parserBase]]
            Dictionary of filename and corresponding parser objects, or a list of parser objects upon which filenames will be interred from the filename attribute.
        load_all_columns : bool
            Load all columns unreferenced by `COLUMN_ASSIGNMENTS` parser object attribute, by default True.
        override_scans : bool
            Override existing scan objects with the same filename, by default False.
        """
        # Convert parsers list to dict.
        if isinstance(parsers, list):
            parsers = {parser.filename: parser for parser in parsers}
        # Update scans with new filename parsers unless override flag is set.
        scans = self.scans
        self.scans = {
            filename: parser.to_scan(load_all_columns)
            for filename, parser in parsers.items()
            if (override_scans or filename not in scans)
        }

    @property
    def scans(self) -> dict[str, type[scanAbstract]]:
        """
        Scan objects loaded into the viewer.

        Returns
        -------
        dict[str, scan_abstract]
            A shallow copy of the dictionary object containing scans.
        """
        return self._scan_objects.copy()

    @scans.setter
    def scans(self, scans: dict[str, scanAbstract]):
        """
        Setter for the scan objects.

        Adds new scan objects and updates existing scan objects with the same filename.

        Parameters
        ----------
        scans : dict[str, scan_abstract]
            A dictionary of filename keys and corresponding scan object values.
        """
        self._scan_objects.update(scans)

    @scans.deleter
    def scans(self):
        """
        Flushes the existing scan objects from the viewer.
        """
        self._scan_objects = {}

    @property
    def selected_filenames(self) -> list[str]:
        """
        The current selection of files. Empty list if None.

        Returns
        -------
        list[str]
            List of filenames selected.
        """
        return self._selected_filenames

    @selected_filenames.setter
    def selected_filenames(self, names: list[str]):
        """
        Setter for the selected files.

        Parameters
        ----------
        names : list[str]
            List of filenames to select.
        """
        # Update internal object
        self._selected_filenames = names.copy() if names is not None else []
        # Update the dataseries list
        self._update_dataseries_list()

    def _update_dataseries_list(self) -> None:
        """
        Update the dataseries list widget with the current selection of scan objects.

        Preserves name selection if updated dataseries contain the same names.
        Alternatively preserves the selected indexes if the names are not present, but the indexes are (such as for relabelling).
        """
        # Prevent the callback from triggering the update if updating.
        self._updating = True
        # Get new filenames
        names = self._selected_filenames
        # Update the data series list
        if len(names) != 0:
            # Store Previous Selection
            previousSelection = self.dataseries_selected
            previousQModelIndexes = (
                self._dataseries_list_view.selectionModel().selectedRows()
            )
            previousIndexes = [index.row() for index in previousQModelIndexes]

            # Find common labels in scan objects:
            scans = self.scans
            # Only consider names that have scan objects.
            names = [name for name in names if name in scans]

            if len(names) > 1:
                # Look through all scan objects
                scan = scans[names[0]]
                labels = (
                    set(scan.y_labels) if scan is not None else []
                )  # set to empty list if scan object is None.
                for name in names[1:]:
                    if name in scans:
                        labels = labels.intersection(set(scans[name].y_labels))
            elif len(names) == 1:
                scan = scans[names[0]]
                labels = scan.y_labels
            else:
                labels = []
                scan = None
            # Update attributes in order, keeping original appearance order.
            ordered_intersection = []
            if scan is not None:
                for label in scan.y_labels:
                    if label in labels:
                        ordered_intersection.append(label)
            else:
                ordered_intersection = []
            self._dataseries_list = ordered_intersection
            # Update graphics
            self._dataseries_list_view.clear()
            self._dataseries_list_view.addItems(self._dataseries_list)

            # Restore the previous selection if it exists.
            if previousSelection is not None:
                # Restore the previous selection if it exists.
                if np.all(
                    [label in ordered_intersection for label in previousSelection]
                ):
                    # Use exact same labels
                    self.dataseries_selected = previousSelection
                elif len(ordered_intersection) > np.max(previousIndexes):
                    # Use indexes of labels
                    # self._dataseries_list_view.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectionBehavior.SelectRows)
                    # self._dataseries_list_view.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.MultiSelection)
                    for i in previousIndexes:
                        self._dataseries_list_view.item(i).setSelected(True)
                    self._dataseries_list_view.setFocus()
        else:
            self._dataseries_list = []
            self._dataseries_list_view.clear()

        # Reset the updating flag.
        self._updating = False

        # Update the normalisation graph.
        # TODO: DEBUG THIS TO ENABLE!
        # self.on_label_selection_change()

    @property
    def dataseries_selected(self) -> list[str] | None:
        """
        The currently selected data series common to all scan objects.

        Returns
        -------
        list[str]
            List of strings of the data series selected.
        """
        selected_items = self._dataseries_list_view.selectedItems()
        if len(selected_items) > 0:
            return [item.text() for item in selected_items]
        else:
            return None

    @dataseries_selected.setter
    def dataseries_selected(self, labels: list[str]):
        """
        Setter for the selected data series.

        Parameters
        ----------
        labels : list[str]
            List of labels to select.
        """
        for i in range(self._dataseries_list_view.count()):
            item = self._dataseries_list_view.item(i)
            if item.text() in labels:
                item.setSelected(True)
            else:
                item.setSelected(False)

    def on_label_selection_change(self) -> None:
        """
        Callback for when the selected labels change.
        """
        # Get the current scan object selection.
        file_selection = self.selected_filenames
        # Prevent the callback from triggering the update if updating.
        if (
            not self._updating
            and file_selection is not None
            and len(file_selection) > 0
        ):
            scans_subset = [
                scan for name, scan in self.scans.items() if name in file_selection
            ]
            # Get the selected fields.
            ds_list = self.dataseries_selected
            # Propogate dataseries selection to normalisation widget.
            self._normGraph.graph_scans = scans_subset
            self._normGraph.dataseries_selection = (
                ds_list if ds_list is not None else []
            )
            self._normGraph.graph_selection()

    def on_normalisation_change(self) -> None:
        """
        Callback for when the normalisation changes.
        """
        self._norm_settings = self._normGraph.norm_settings

    def on_relabel(self) -> None:
        """
        Callback for when the labels are relabelled.
        """
        for scan in self.scans.values():
            if isinstance(scan, scanAbstract):
                assert isinstance(scan, scanAbstract)
                scan.reload_labels_from_parser()

        self._update_dataseries_list()


class normalisingGraph(QWidget):
    def __init__(
        self,
        graph_scans: list[Type[scanAbstract | parserBase]] = None,
        dataseries_selection: list[str] = [],
        background_fixed_scans: list[Type[scanAbstract]] = [],
        # norm_settings: normaliserSettings = None,
        parent=None,
    ):
        super().__init__(parent)
        self._layout = QVBoxLayout()
        self.setLayout(self._layout)
        if parent is not None:
            self.setContentsMargins(0, 0, 0, 0)
            self._layout.setContentsMargins(0, 0, 0, 0)
        self._figure = matplotlib.figure.Figure()
        self._canvas = FigureCanvas(self._figure)  # use empty canvas.
        self._toolbar = NavTBQT(self._canvas)
        # self._toolbar = NavTBQT_Norm(self._canvas)
        self._layout.addWidget(self._toolbar)
        self._layout.addWidget(self._canvas)

        # Update toolbar settings.
        self.graph_scans = graph_scans
        self.dataseries_selection = dataseries_selection
        self.background_fixed_scans = background_fixed_scans
        # self.norm_settings = norm_settings

    @property
    def graph_scans(self) -> list[Type[scanAbstract | parserBase]]:
        return self._toolbar.graph_scans.copy()

    @graph_scans.setter
    def graph_scans(self, scans: list[Type[scanAbstract | parserBase]]):
        self.toolbar.graph_scans = scans.copy() if scans is not None else None

    @property
    def dataseries_selection(self) -> list[str]:
        return self.toolbar.dataseries_selection.copy()

    @dataseries_selection.setter
    def dataseries_selection(self, labels: list[str]):
        self.toolbar.dataseries_selection = (
            labels.copy() if labels is not None else None
        )

    @property
    def background_fixed_scans(self) -> list[Type[scanAbstract | parserBase]]:
        return self._toolbar.background_fixed_scans.copy()

    @background_fixed_scans.setter
    def background_fixed_scans(self, scans: list[Type[scanAbstract | parserBase]]):
        self.toolbar.background_fixed_scans = (
            scans.copy() if scans is not None else None
        )

    # @property
    # def norm_settings(self) -> normaliserSettings:
    #     return self._toolbar.norm_settings

    # @norm_settings.setter
    # def norm_settings(self, settings: normaliserSettings):
    #     self.toolbar.norm_settings = settings if settings is not None else None

    @property
    def figure(self) -> matplotlib.figure.Figure:
        return self._figure

    @figure.setter
    def figure(self, fig: matplotlib.figure.Figure):
        self._figure = fig
        self._canvas.figure = fig

    @property
    def canvas(self) -> FigureCanvas:
        return self._canvas

    @property
    def toolbar(self):
        return self._toolbar

    def graph_selection(self) -> None:
        scans = self.graph_scans
        dataseries_list = self.dataseries_selection
        self.figure.clear()
        ax = self.figure.add_subplot(111)
        assert isinstance(ax, matplotlib.axes.Axes)
        # Iterate over dataseries first:
        if dataseries_list is not None and len(scans) > 0:
            scan = scans[0]
            for ds in dataseries_list:
                # Then iterate over each scan | parser object.
                for scan in scans:
                    try:
                        if isinstance(scan, scanAbstract):
                            ind = scan.y_labels.index(ds)
                            x = scan.x
                            y = scan.y[:, ind]
                            ax.plot(x, y, label=scan.filename + ":" + ds)
                        elif isinstance(scan, parserBase):
                            assert isinstance(scan, parserBase)

                    except ValueError:
                        # Catch Value errors if the label is not found: don't plot.
                        continue
            ax.set_xlabel(
                scan.x_label + " (" + scan.x_unit + ")"
                if scan.x_unit is not None
                else scan.x_label
            )
            # Copy behaviour from matplotlib figureoptions.py to allow removal of legend with user modification.
            draggable = None
            ncols = 1
            if ax.legend_ is not None:
                old_legend = ax.get_legend()
                draggable = old_legend._draggable is not None
                ncols = old_legend._ncols
            new_legend = ax.legend(ncols=ncols)
            if new_legend:
                new_legend.set_draggable(draggable)
        self.canvas.draw()


if __name__ == "__main__":
    app = QApplication(sys.argv)

    window = nexafsViewer()
    window.setWindowTitle("pyNexafs nexafsViewer")
    # window = normalisingGraph()
    # window.setWindowTitle("pyNexafs Normalising Graph")
    window.show()
    sys.exit(app.exec())



from PyQt6 import QtWidgets, QtCore, QtGui
from pyNexafs.parsers._base import parserBase
from pyNexafs.nexafs.scan import scanBase
from typing import Type
from builtins import function
import sys
from pyNexafs.gui.widgets.io.dir_selection import directorySelector
import pyNexafs.parsers.au as AU_PARSERS
from pyNexafs.gui.widgets.graphing.matplotlib.graphs import FigureCanvas, NEXAFS_NavQT
import warnings
from enum import Enum
import numpy as np
import numpy.typing as npt
import os
import matplotlib.pyplot as plt


class save_directory_selector(directorySelector):
    edit_description = "Path to save QANT-compatible data to."
    dialog_caption = "Select Save Directory"

    def __init__(self, parent=None):
        super().__init__(parent)
        layout: QtWidgets.QHBoxLayout = self.layout()
        layout.insertWidget(0, QtWidgets.QLabel("Save Directory:"))


class NaN_OPTION(Enum):
    """
    Enumerate to define methods of processing NaN (Not a Number) values.

    Attributes
    ----------
    NONE : int
        No special treatment. Data stored with NaN labels. Incompatible with QANT?
    REMOVE : int
        Removes NaN rows. Affects multiple data columns that might not have NaNs.
    INTERPOLATE_2 : int
        Interpolates the nearest 2 (non-NaN) point neighbours to
    """

    NONE = 0
    REMOVE_ROWS = 1
    INTERPOLATE_2_POINTS = 2
    INTERP_EXTRAP_2_POINTS = 3
    # INTERPOLATE_4_POINTS = 3
    # INTERPOLATE_N_POINTS = 4


class horLine(QtWidgets.QFrame):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFrameShape(QtWidgets.QFrame.Shape.HLine)
        self.setFrameShadow(QtWidgets.QFrame.Shadow.Sunken)
        self.setLineWidth(1)
        self.setMinimumHeight(1)
        self.setMinimumWidth(1)
        self.setSizePolicy(
            QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Minimum
        )

    def event(self, event: QtCore.QEvent) -> bool:
        """
        Event handler for the widget.

        Adds palette change control for light/dark mode to QWidget event handler.

        Parameters
        ----------
        event : QtCore.QEvent
            The event to handle.

        Returns
        -------
        bool
            Whether the event was handled.
        """
        if (
            event.type() == QtCore.QEvent.Type.PaletteChange
            or event.type() == QtCore.QEvent.Type.ApplicationPaletteChange
        ):
            self.on_recolour()
        return super().event(event)

    def on_recolour(self):
        """
        Recolour the division lines based on the theme.
        """
        # Get theme
        toolbar_palette = self.palette()
        light_theme_bool = toolbar_palette.window().color().lightnessF() > 0.5
        self.setStyleSheet(
            "background-color: " + ("black;" if light_theme_bool else "white;")
        )


class nexafsParserConverter(QtWidgets.QWidget):
    def __init__(self, parsers: list[Type[parserBase | scanBase]] = [], parent=None):
        super().__init__(parent)
        self._parsers = {parser.filename: parser for parser in parsers}
        self._conversions = []
        self._layout = QtWidgets.QVBoxLayout()
        self.setLayout(self._layout)
        if parent is not None:
            self.setContentsMargins(0, 0, 0, 0)
            self._layout.setContentsMargins(0, 0, 0, 0)

        ## Setup UI elements
        # Title
        self._title_label = title_label = QtWidgets.QLabel("Converstion:")
        # Folder selector
        self.dir_sel = dir_sel = save_directory_selector()

        # File statistics
        file_widget = QtWidgets.QWidget()
        file_layout = QtWidgets.QGridLayout()
        file_label = QtWidgets.QLabel("File Statistics")
        file_layout.addWidget(file_label, 0, 0, 1, 4)
        file_widget.setLayout(file_layout)
        file_widget.setContentsMargins(0, 0, 0, 0)
        file_layout.setContentsMargins(0, 0, 0, 0)
        # Number of files
        files_num_label = QtWidgets.QLabel("Number of Files:")
        self.files_num_indicator = QtWidgets.QLabel("")
        file_layout.addWidget(files_num_label, 1, 0)
        file_layout.addWidget(self.files_num_indicator, 1, 1)
        # Total filesize
        files_size_label = QtWidgets.QLabel("Total Filesize:")
        self.files_size_indicator = QtWidgets.QLabel("")
        file_layout.addWidget(files_size_label, 1, 2)
        file_layout.addWidget(self.files_size_indicator, 1, 3)

        # # NaN inspector
        nan_widget = QtWidgets.QWidget()
        nan_layout = QtWidgets.QGridLayout()
        nan_label = QtWidgets.QLabel("NaN Inspector:")
        nan_layout.addWidget(nan_label, 0, 0, 1, 2)
        nan_widget.setLayout(nan_layout)
        nan_widget.setContentsMargins(0, 0, 0, 0)
        nan_layout.setContentsMargins(0, 0, 0, 0)
        # Number of files with NaNs
        total_nan_files_label = QtWidgets.QLabel("Files with NaNs:")
        self.total_nan_files_indicator = QtWidgets.QLabel("")
        nan_layout.addWidget(total_nan_files_label, 0, 2)
        nan_layout.addWidget(self.total_nan_files_indicator, 0, 3)
        # Total nan indicator:
        total_nan_instances_label = QtWidgets.QLabel("Total NaN Instances:")
        self.total_nan_instances_indicator = QtWidgets.QLabel("")
        nan_layout.addWidget(total_nan_instances_label, 1, 0)
        nan_layout.addWidget(self.total_nan_instances_indicator, 1, 1)
        # Per file nan indicator:
        per_file_nan_label = QtWidgets.QLabel("NaN Instances per Nan File:")
        self.per_file_nan_indicator = QtWidgets.QLabel("")
        nan_layout.addWidget(per_file_nan_label, 1, 2)
        nan_layout.addWidget(self.per_file_nan_indicator, 1, 3)
        # Method of dealing with Nan Values.
        nan_method_label = QtWidgets.QLabel("NaN Method:")
        self.nan_method = QtWidgets.QComboBox()
        self.nan_method.addItems(
            [
                key.replace("_", " ").lower().capitalize()
                for key in NaN_OPTION.__members__.keys()
            ]
        )
        nan_layout.addWidget(nan_method_label, 2, 0, 1, 1)
        nan_layout.addWidget(self.nan_method, 2, 1, 1, 3)

        # Number of files with Zeros:
        total_zero_files_label = QtWidgets.QLabel("Files with Zero Blocks:")
        self.total_zero_files_indicator = QtWidgets.QLabel("")
        nan_layout.addWidget(total_zero_files_label, 3, 2)
        nan_layout.addWidget(self.total_zero_files_indicator, 3, 3)
        # Total zeros indicator:
        total_zero_instances_label = QtWidgets.QLabel("Total Zero Instances:")
        self.total_zero_instances_indicator = QtWidgets.QLabel("")
        nan_layout.addWidget(total_zero_instances_label, 4, 0)
        nan_layout.addWidget(self.total_zero_instances_indicator, 4, 1)
        # Per file zeros indicator:
        per_file_zero_label = QtWidgets.QLabel("Zero Instances per Zero Block File:")
        self.per_file_zero_indicator = QtWidgets.QLabel("")
        nan_layout.addWidget(per_file_zero_label, 4, 2)
        nan_layout.addWidget(self.per_file_zero_indicator, 4, 3)
        # Treat block zeros as NaNs
        self.nan_zero_checkbox = QtWidgets.QCheckBox("Block Zeros as NaN?")
        self.nan_zero_checkbox.setToolTip(
            "Treat blocks of zeros in data as NaN values."
        )
        nan_layout.addWidget(self.nan_zero_checkbox, 3, 0, 1, 2)

        ## Converter Buttons / Options
        exec_widget = QtWidgets.QWidget()
        exec_layout = QtWidgets.QGridLayout()
        exec_widget.setLayout(exec_layout)
        exec_widget.setContentsMargins(0, 0, 0, 0)
        exec_layout.setContentsMargins(0, 0, 0, 0)
        conversion_text = QtWidgets.QLabel("Conversion Execution:")
        self.progress = progress = QtWidgets.QProgressBar()
        progress.setMinimumHeight(5)
        progress.setMinimumWidth(200)
        progress.setTextVisible(False)
        progress.setOrientation(QtCore.Qt.Orientation.Horizontal)
        progress.setValue(0)
        copy_button = QtWidgets.QPushButton("Copy (single) to Clipboard")
        save_button = QtWidgets.QPushButton("Save (selection) to File")
        self._override_checkbox = override_checkbox = QtWidgets.QCheckBox(
            "Override Existing Files?"
        )
        self._rename_checkbox = rename_checkbox = QtWidgets.QCheckBox(
            "Rename? (`<sample>_<scan_id>.asc`)"
        )
        exec_layout.addWidget(conversion_text, 0, 0, 1, 2)
        exec_layout.addWidget(copy_button, 0, 2, 1, 2)
        exec_layout.addWidget(override_checkbox, 1, 0, 1, 2)
        exec_layout.addWidget(rename_checkbox, 1, 2, 1, 2)
        exec_layout.addWidget(progress, 2, 0, 1, 4)
        exec_layout.addWidget(save_button, 3, 0, 1, 4)

        ## Difference viewer
        diff_widget = QtWidgets.QWidget()
        diff_layout = QtWidgets.QGridLayout()
        diff_widget.setLayout(diff_layout)
        diff_widget.setContentsMargins(0, 0, 0, 0)
        diff_layout.setContentsMargins(0, 0, 0, 0)
        diff_label = QtWidgets.QLabel("Differences:")
        diff_layout.addWidget(diff_label, 0, 1, 1, 3)
        self._selected_parser = None
        file_label = QtWidgets.QLabel("File:")
        self.diff_parser_selector = QtWidgets.QComboBox()
        series_label = QtWidgets.QLabel("Series:")
        self.diff_series_selector = QtWidgets.QComboBox()
        diff_layout.addWidget(file_label, 1, 0, 1, 3)
        diff_layout.addWidget(self.diff_parser_selector, 1, 1, 1, 3)
        diff_layout.addWidget(series_label, 2, 0, 1, 3)
        diff_layout.addWidget(self.diff_series_selector, 2, 1, 1, 3)
        self.fig = fig = plt.figure(figsize=(2, 2))
        self.diff_canvas = FigureCanvas(fig)
        self.diff_tb = NEXAFS_NavQT(self.diff_canvas, self)
        diff_layout.addWidget(self.diff_canvas, 3, 0, 1, 4)
        diff_layout.addWidget(self.diff_tb, 4, 0, 1, 4)

        # Track nan_parsers
        self._nan_parsers = {}

        # Add to layout
        self._layout.addWidget(title_label)
        self._layout.addWidget(dir_sel)
        self._layout.addWidget(horLine())
        self._layout.addWidget(file_widget)
        self._layout.addWidget(horLine())
        self._layout.addWidget(nan_widget)
        self._layout.addWidget(horLine())
        self._layout.addWidget(exec_widget)
        self._layout.addWidget(horLine())
        self._layout.addWidget(diff_widget)
        self._layout.setAlignment(QtCore.Qt.AlignmentFlag.AlignTop)

        # Connect signals
        copy_button.clicked.connect(self.on_copy_to_clipboard)
        save_button.clicked.connect(self.on_save_to_file)
        self.nan_method.currentIndexChanged.connect(self.on_conversion_change)
        self.nan_zero_checkbox.stateChanged.connect(self.on_conversion_change)
        self.diff_parser_selector.currentIndexChanged.connect(
            self.on_diff_parser_change
        )
        self.diff_series_selector.currentIndexChanged.connect(
            self.on_diff_series_change
        )

    def on_conversion_change(self):
        # Recalculate the conversions.
        self.calculate_nan()
        self.graph_diff()

    def on_parsers_selected(self):
        """Updates the parser selector based on the provided parsers objects."""
        self.diff_parser_selector.clear()
        self._nan_parsers.clear()
        current_idx = 0  # Used to keep track of ComboBox index.
        for i, parser in enumerate(self.parsers):
            if np.any(np.isnan(parser.data)):
                self.diff_parser_selector.addItem(parser.filename)
                # Add a mapping to
                self._nan_parsers[current_idx] = i
                current_idx += 1
        # Call the diff parser change to update the series selector.
        self.on_diff_parser_change()

    def on_diff_parser_change(self):
        """Updates the series selector based on the selected parser."""
        self.diff_series_selector.blockSignals(True)
        self.diff_series_selector.clear()
        current_idx = self.diff_parser_selector.currentIndex()
        parsers = self.parsers
        if current_idx is None:
            self.diff_series_selector.clear()
            return
        elif (
            isinstance(current_idx, int)
            and current_idx > -1
            and len(parsers) > current_idx
        ):
            parser = parsers[self._nan_parsers[current_idx]]
            for i in range(parser.data.shape[1]):
                if np.any(np.isnan(parser.data[:, i])):
                    self.diff_series_selector.addItem(parser.labels[i])
            self.on_diff_series_change()
        self.diff_series_selector.blockSignals(False)
        # Call the series change.
        self.on_diff_series_change()

    def on_diff_series_change(self):
        """Actions to perform when a parser/series combination is selected."""
        self.graph_diff()

    def graph_diff(self):
        """Graphs the raw and nan-processed data of the selected parser/series."""
        selector_index = self.diff_parser_selector.currentIndex()
        self.fig.clear()
        # 0 is the empty string
        if selector_index is not None and selector_index > -1:
            parser_idx = self._nan_parsers[selector_index]
            # Collect parser
            current_parser = self.parsers[parser_idx]
            current_converted = self._conversions[parser_idx]
            # Collect the index of the nan data series
            series_idx = current_parser.labels.index(
                self.diff_series_selector.currentText()
            )

            # Graph both converted and original data
            fig = self.fig
            ax = self.fig.add_subplot(111)
            x_idx = current_parser.label_index(current_parser.COLUMN_ASSIGNMENTS["x"])

            # Plot the converted data depending on the NaN method.
            nan_method = NaN_OPTION(self.nan_method.currentIndex())
            match nan_method:
                # # Locate the nan domain
                # parser_nans = np.isnan(current_parser.data[:, series_idx])
                # # Let nearest neighbours also be displayed in the nan data:
                # roll_pos = np.roll(parser_nans, 1)
                # roll_neg = np.roll(parser_nans, -1)
                # parser_nans = roll_neg | roll_pos | parser_nans
                case NaN_OPTION.REMOVE_ROWS:
                    # Plot raw data after, so gaps can be seen underneath.
                    ax.plot(
                        current_converted.data[:, x_idx],
                        current_converted.data[:, series_idx],
                        label="NaN Fixed",
                        c="r",
                    )
                    ax.plot(
                        current_parser.data[:, x_idx],
                        current_parser.data[:, series_idx],
                        label="Original",
                        c="b",
                    )
                case NaN_OPTION.INTERPOLATE_2_POINTS:
                    # Plot raw data after, so gaps can be seen underneath.
                    ax.plot(
                        current_converted.data[:, x_idx],
                        current_converted.data[:, series_idx],
                        label="NaN Fixed",
                        c="r",
                    )
                    ax.plot(
                        current_parser.data[:, x_idx],
                        current_parser.data[:, series_idx],
                        label="Original",
                        c="b",
                    )
                case NaN_OPTION.INTERP_EXTRAP_2_POINTS:
                    # Plot raw data after, so gaps can be seen underneath.
                    ax.plot(
                        current_converted.data[:, x_idx],
                        current_converted.data[:, series_idx],
                        label="NaN Fixed",
                        c="r",
                    )
                    ax.plot(
                        current_parser.data[:, x_idx],
                        current_parser.data[:, series_idx],
                        label="Original",
                        c="b",
                    )
                    pass
                case _:
                    ax.plot(
                        current_converted.data[:, x_idx],
                        current_converted.data[:, series_idx],
                        label="Data",
                        c="b",
                    )

            ax.set_xlabel(current_parser.labels[x_idx])
            ax.set_ylabel(current_parser.labels[series_idx])
            ax.legend()
            fig.tight_layout()
            self.diff_canvas.draw()

    def on_copy_to_clipboard(self):
        if len(self._conversions) > 0:
            if len(self._conversions) != 1:
                warnings.warn(
                    "More than one conversion performed. Copying first conversion.",
                    stacklevel=2,
                )
            # Run the conversion on the first file.
            parser_type = type(self.parsers[0])
            match parser_type:
                case AU_PARSERS.MEX1_NEXAFS:
                    lines = AU_PARSERS.MEX1_to_QANT_AUMainAsc(self._conversions[0])
                case AU_PARSERS.MEX2_NEXAFS:
                    lines = AU_PARSERS.MEX2_to_QANT_AUMainAsc(self._conversions[0])
                case _:
                    raise NotImplementedError(
                        f"Conversion for this parser type {type(self.parsers[0])} has not been implemented."
                    )
            cb = QtGui.QGuiApplication.clipboard()
            cb.setText("".join(lines), mode=QtGui.QClipboard.Mode.Clipboard)
        else:
            # Do nothing
            pass

    def on_save_to_file(self):
        base_dir = self.dir_sel._folder_path
        init_override = False
        if len(self._conversions) > 0:
            self.progress.setRange(0, len(self._conversions))
            for parser in self._conversions:
                ptype = type(parser)
                converter: function
                match ptype:
                    case AU_PARSERS.MEX1_NEXAFS:
                        converter = AU_PARSERS.MEX1_to_QANT_AUMainAsc
                    case AU_PARSERS.MEX2_NEXAFS:
                        converter = AU_PARSERS.MEX2_to_QANT_AUMainAsc
                self.progress.setValue(self.progress.value() + 1)
                # Convert to ascii format.
                if self._rename_checkbox.isChecked() and (
                    "SR14ID01NEXSCAN:saveData_comment1" in parser.params
                    or "Comment 1" in parser.params
                    or "Sample" in parser.params
                ):
                    # Get the sample name
                    if "SR14ID01NEXSCAN:saveData_comment1" in parser.params:
                        sample = parser.params["SR14ID01NEXSCAN:saveData_comment1"]
                    elif "Comment 1" in parser.params:
                        sample = parser.params["Comment 1"]
                    elif "Sample" in parser.params:
                        sample = parser.params["Sample"]
                    else:
                        print(f"Name error, skipping {parser}")
                        continue

                    # Get the scan ID / number
                    if "mda_scan_number" in parser.params:
                        scan_id = parser.params["mda_scan_number"]
                    elif "Scan Number" in parser.params:
                        scan_id = parser.params["Scan Number"]
                    else:
                        # use the filename?
                        scan_id = "".join(parser.filename.split(".")[:-1])
                    # Create the new filename
                    file = f"{sample}_{scan_id}.asc"
                else:
                    file = "".join(parser.filename.split(".")[:-1]) + ".asc"
                path = os.path.join(base_dir, file)
                exists = os.path.exists(path)
                if self._override_checkbox.isChecked() or not exists:
                    if init_override is False and exists:
                        diag = QtWidgets.QMessageBox(self)
                        diag.setWindowTitle("Files Exist")
                        diag.setText(
                            f"File {file} already exists. Override?\n\nPath: {path}"
                        )
                        diag.setStandardButtons(
                            QtWidgets.QMessageBox.StandardButton.YesToAll
                            | QtWidgets.QMessageBox.StandardButton.Yes
                            | QtWidgets.QMessageBox.StandardButton.No
                        )
                        response = diag.exec()
                        match response:
                            case QtWidgets.QMessageBox.StandardButton.YesToAll:
                                init_override = True
                            case QtWidgets.QMessageBox.StandardButton.Yes:
                                with open(path, "w") as f:
                                    f.write("".join(converter(parser)))
                            case _:
                                pass
                    else:
                        with open(path, "w") as f:
                            f.write("".join(converter(parser)))
                elif os.path.exists(path):
                    diag = QtWidgets.QMessageBox(self)
                    diag.setWindowTitle("Files Exist")
                    diag.setText(
                        f"File {file} already exists.\nSelect checkbox to overwrite."
                    )
            self.progress.setValue(0)

    def calculate_nan(self, single_file=False) -> None:
        """
        Calculate the NaN conversions of the data.
        """
        parsers = self.parsers
        method = NaN_OPTION(self.nan_method.currentIndex())
        zero_as_nan = self.nan_zero_checkbox.isChecked()
        if single_file:
            self.progress.setRange(0, 1)
            self.progress.setValue(0)
            self._conversions = [
                nexafsParserConverter.calculation_single(
                    parsers[0], method, zero_as_nan
                )
            ]
        elif len(parsers) > 0:
            self._conversions = []
            self.progress.setRange(0, len(parsers))
            self.progress.setValue(0)
            for i, parser in enumerate(parsers):
                # Update progress bar
                self.progress.setValue(i + 1)
                copy_parser = nexafsParserConverter.calculation_single(
                    parser, method, zero_as_nan
                )
                self._conversions.append(copy_parser)
        # Finished calculating.
        self.progress.setValue(0)

    @staticmethod
    def calculation_single(
        parser: parserBase | scanBase,
        method: NaN_OPTION,
        treat_zero_as_nan: bool = False,
        zero_block_min_size=3,
    ) -> parserBase | scanBase:
        # Copy parser
        copy_parser = parser.copy()
        # Treat zeros as NaN values
        if treat_zero_as_nan:
            blocks = nexafsParserConverter._get_block_zeros(
                data=copy_parser.data, zero_block_min_size=zero_block_min_size
            )
            # Set blocks to NaN
            for i, j, width, height in blocks:
                copy_parser.data[i : i + width, j : j + height] = np.nan

        # Trim nan values from beginning and end of data - leave values in between.
        nan_vals = np.isnan(copy_parser.data)
        row_nan_vals = np.all(nan_vals, axis=1)
        # Find the first non-nan value
        start_idx = np.argmax(~row_nan_vals)
        # Find the last non-nan value
        end_idx = len(row_nan_vals) - np.argmax(~row_nan_vals[::-1])
        # Trim the data
        copy_parser.data = copy_parser.data[start_idx:end_idx]

        # Act on NaN values
        x_idx = parser.label_index(parser.COLUMN_ASSIGNMENTS["x"])
        match method:
            case NaN_OPTION.REMOVE_ROWS:
                # Remove data rows that contain nans:
                nan_idx = np.isnan(copy_parser.data).any(axis=1)
                # For row, delete if any nans are present.
                copy_parser.data = np.delete(copy_parser.data, nan_idx, axis=0)

            case NaN_OPTION.INTERPOLATE_2_POINTS:
                # Interpolate the nearest 2 points to fill in NaN values.
                for j, column in enumerate(copy_parser.data.T):
                    nan_values = np.isnan(column)
                    if nan_values.any():
                        if not (True in nan_values and False in nan_values):
                            # All NaN values, skip.
                            continue
                        # Uses linear interpolation to fill in NaN values.
                        if j == x_idx:
                            # Assume equal separation between x values
                            copy_parser.data[:, j][nan_values] = np.interp(
                                x=np.arange(len(copy_parser.data))[nan_values],
                                xp=np.arange(len(copy_parser.data))[~nan_values],
                                fp=copy_parser.data[:, j][~nan_values],
                            )
                            continue
                        else:
                            copy_parser.data[:, j][nan_values] = np.interp(
                                x=copy_parser.data[:, x_idx][nan_values],
                                xp=copy_parser.data[:, x_idx][~nan_values],
                                fp=copy_parser.data[:, j][~nan_values],
                            )
            case NaN_OPTION.INTERP_EXTRAP_2_POINTS:
                ### Extrapolate and interpolate the nearest 2 points to fill in NaN values.
                start = []
                end = []
                # Interpolation
                for j, column in enumerate(copy_parser.data.T):
                    nan_values = np.isnan(column)
                    ## Find column endpoints for extrapolation
                    # Find the first non-nan value
                    start_idx = np.argmax(~nan_values)
                    start.append(start_idx)
                    # Find the last non-nan value
                    end_idx = len(nan_values) - np.argmax(~nan_values[::-1])
                    end.append(end_idx)
                    # Uses linear interpolation to fill in NaN values.
                    if nan_values.any():
                        if j == x_idx:
                            # Assume equal separation between x values
                            copy_parser.data[:, j][nan_values] = np.interp(
                                x=np.arange(len(copy_parser.data))[nan_values],
                                xp=np.arange(len(copy_parser.data))[~nan_values],
                                fp=copy_parser.data[:, j][~nan_values],
                            )
                            continue
                        else:
                            # Calculate the average delta between x values
                            xdiff = np.abs(
                                np.diff(copy_parser.data[:, x_idx][~nan_values])
                            )
                            delta = np.mean(xdiff)
                            # Interpolate x data
                            copy_parser.data[:, j][nan_values] = np.interp(
                                x=copy_parser.data[:, x_idx][nan_values],
                                xp=copy_parser.data[:, x_idx][~nan_values],
                                fp=copy_parser.data[:, j][~nan_values],
                            )

                ### Extrapolation
                # Do x first...
                if j == x_idx and xdiff.std() < delta / 30:
                    # Assume data density consistent and delta can be used for x extrapolation
                    if start[x_idx] > 0:
                        points = start[x_idx]
                        first_val = copy_parser.data[start[x_idx], x_idx]
                        # increasing or decreasing?
                        sign = (
                            -1
                            if first_val > copy_parser.data[start[x_idx] + 1, x_idx]
                            else 1
                        )
                        copy_parser.data[: start[x_idx], x_idx] = np.linspace(
                            start=first_val - delta * points * sign,
                            stop=first_val - delta * sign,
                            num=points,
                        )
                    if end[x_idx] < len(copy_parser.data):
                        last_val = copy_parser.data[end[x_idx], x_idx]
                        copy_parser.data[end[x_idx] :, x_idx] = np.linspace(
                            start=last_val + delta,
                            stop=last_val
                            + delta * (len(copy_parser.data) - end[x_idx]),
                            num=len(copy_parser.data) - end[x_idx],
                        )
                else:
                    if len(copy_parser.data[start[x_idx] : end[x_idx]]) > 1:
                        # Start
                        y1 = copy_parser.data[start[x_idx] + 1, x_idx]
                        y2 = copy_parser.data[start[x_idx], x_idx]
                        sgrad = (y2 - y1) / 2
                        n = start[x_idx]
                        copy_parser.data[: start[x_idx], x_idx] = np.linspace(
                            start=copy_parser.data[start[x_idx], x_idx] + sgrad,
                            stop=copy_parser.data[start[x_idx], x_idx] + sgrad * n,
                            num=n,
                        )
                        # End
                        y1 = copy_parser.data[end[x_idx] - 2, x_idx]
                        y2 = copy_parser.data[end[x_idx] - 1, x_idx]
                        sgrad = (y2 - y1) / 2
                        n = len(copy_parser.data) - end[x_idx]
                        copy_parser.data[end[x_idx] :, x_idx] = np.linspace(
                            start=copy_parser.data[end[x_idx] - 1, x_idx] + sgrad,
                            stop=copy_parser.data[end[x_idx] - 1, x_idx] + sgrad * n,
                            num=n,
                        )

                # Now do other columns
                for j in range(len(start)):
                    if j == x_idx:
                        continue
                    else:
                        start_x = copy_parser.data[: start[j], x_idx]
                        end_x = copy_parser.data[end[j] :, x_idx]
                        # Extrapolate everything closest last two points.
                        if len(copy_parser.data[start[j] : end[j]]) > 1:
                            ## Valid range to extrapolate.
                            # Start
                            y1 = copy_parser.data[start[j] + 1, j]
                            y2 = copy_parser.data[start[j], j]
                            sgrad = (y2 - y1) / 2
                            n = start[j]
                            copy_parser.data[: start[j], j] = np.interp(
                                x=start_x,
                                xp=copy_parser.data[start[j] : start[j] + 2, x_idx],
                                fp=copy_parser.data[start[j] : start[j] + 2, j],
                            )
                            # End
                            # Valid range to extrapolate.
                            y1 = copy_parser.data[end[j] - 1, j]
                            y2 = copy_parser.data[end[j] - 2, j]
                            sgrad = (y2 - y1) / 2
                            n = len(copy_parser.data) - end[j]
                            copy_parser.data[end[j] :, j] = np.interp(
                                x=end_x,
                                xp=copy_parser.data[
                                    end[j] - 1 : end[j] - 3 : -1, x_idx
                                ],
                                fp=copy_parser.data[end[j] - 1 : end[j] - 3 : -1, j],
                            )
                        else:
                            # Do nothing, not enough values to extrapolate.
                            pass

            # case NaN_OPTION.INTERPOLATE_4:
            # Uses polynomial interpolation to fill in NaN values.
            # pass
            # case NaN_OPTION.INTERPOLATE_N:
            # pass
            case _:
                # Do nothing for NaN_OPTION.NONE
                pass
        return copy_parser

    def analyse_nan(self):
        """
        Calculate the number of NaN values in the data.
        """
        parsers = self.parsers
        if len(parsers) > 0:
            total_nan_values = 0
            nanfiles = 0
            for parser in parsers:
                nans = np.isnan(parser.data).sum()
                nanfiles += nans > 0
                total_nan_values += nans
            self.total_nan_files_indicator.setText(str(nanfiles))
            self.total_nan_instances_indicator.setText(str(total_nan_values))
            self.per_file_nan_indicator.setText(
                f"{total_nan_values / nanfiles:.2f}" if nanfiles > 0 else "0"
            )
        else:
            self.total_nan_files_indicator.setText("")
            self.total_nan_instances_indicator.setText("")
            self.per_file_nan_indicator.setText("")

    @staticmethod
    def _get_block_zeros(
        data: npt.NDArray, zero_block_min_size: int = 3
    ) -> list[tuple[int, int, int, int]]:
        """
        Collects the blocks of zeros in a 2D Numpy Array.

        While each block found may not be the largest possible block,
        the list will provide full coverage of the data.

        Parameters
        ----------
        data : npt.NDArray
            Data to find blocks of zeros in.
        zero_block_min_size : int, optional
            Minimum 1D block size, by default 3

        Returns
        -------
        list[tuple[int, int, int, int]]
            _description_
        """
        # Get zero locations
        zero_row, zero_col = np.where(data == 0)
        # Collect blocks
        blocks = []  # Filled with i,j,length,width tuples.
        for idx in range(len(zero_row)):
            i, j = zero_row[idx], zero_col[idx]
            width, height = 1, 1
            # Check for vertical blocks
            for r in range(idx + 1, len(zero_row)):
                # Check next row, plus any existing column
                if zero_row[r] == i + width and (
                    zero_col[r] >= j or zero_col[r] < j + height
                ):
                    width += 1
                # Check next column, plus any existing row
                if zero_col[r] == j + height and (
                    zero_row[r] >= i or zero_row[r] < i + width
                ):
                    height += 1
            # If the block is larger than 1x1, add to block list.
            if width > zero_block_min_size or height > zero_block_min_size:
                blocks.append((i, j, width, height))
        return blocks

    def analyse_block_zeros(self):
        parsers = self.parsers
        self.progress.setRange(0, len(parsers))
        zero_files = 0
        total_zeros = 0
        for i, parser in enumerate(parsers):
            self.progress.setValue(i)
            blocks = nexafsParserConverter._get_block_zeros(parser.data)
            # Update the zero files count
            if len(blocks) > 0:
                zero_files += 1
                # Update the zero instances count
                mask = np.zeros_like(parser.data, dtype=bool)
                for i, j, width, height in blocks:
                    mask[i : i + width, j : j + height] = True
                zeros = np.sum(mask)
            else:
                zeros = 0
            total_zeros += zeros
        # Update the zero indicators
        self.total_zero_files_indicator.setText(str(zero_files))
        self.total_zero_instances_indicator.setText(str(total_zeros))
        self.per_file_zero_indicator.setText(
            f"{total_zeros / zero_files:.2f}" if zero_files > 0 else "0"
        )

    def analyse_filesize(self) -> None:
        """
        Analyses the total filesize of the data.
        """
        self.files_num_indicator.setText(str(self.num_files))
        total_filesize = sum([parser.filesize for parser in self.parsers])
        i = 0
        while total_filesize > 1024:
            total_filesize /= 1024
            i += 1
        total_filesize = (
            f"{total_filesize:.2f} {['B', 'KB', 'MB', 'GB', 'TB', 'PB'][i]}"
        )
        self.files_size_indicator.setText(total_filesize)

    def analyse_parsers(self):
        """
        Runs analysis on the provided parsers, updating GUI elements.
        """
        self.analyse_nan()
        self.analyse_block_zeros()
        self.analyse_filesize()

    @property
    def parsers(self) -> list[Type[parserBase] | parserBase]:
        """
        The current set of parsers to use for conversion.

        Parameters
        ----------
        parsers : list[Type[parserBase]]
            A list of parser classes to use for conversion.
            None values are disregarded when setting.

        Returns
        -------
        list[Type[parserBase]]
            A list of parser classes to use for conversion.
        """
        return list(self._parsers.values())

    @parsers.setter
    def parsers(self, parsers: list[Type[parserBase] | parserBase]):
        self._parsers = {
            parser.filename: parser for parser in parsers if parser is not None
        }
        # Analysis the files
        self.analyse_parsers()
        # Calculates the NaN properties
        self.calculate_nan()
        # Updates UI
        self.diff_parser_selector.blockSignals(True)
        self.on_parsers_selected()
        self.diff_parser_selector.blockSignals(False)

    @property
    def conversions(self) -> list[Type[parserBase]]:
        return self._conversions

    @property
    def num_files(self):
        """
        Calculate the number of files in the data.
        """
        return len(self.parsers)

    # def on_recolour(self):
    #     """
    #     Recolour the division lines based on the theme.
    #     """
    #     # Get theme
    #     toolbar_palette = self.palette()
    #     light_theme_bool = toolbar_palette.window().color().lightnessF() > 0.5
    #     self.setStyleSheet(
    #         "background-color: " + ("black;" if light_theme_bool else "white;")
    #     )

    def event(self, event: QtCore.QEvent) -> bool:
        """
        Event handler for the widget.

        Adds palette change control for light/dark mode to QWidget event handler.

        Parameters
        ----------
        event : QtCore.QEvent
            The event to handle.

        Returns
        -------
        bool
            Whether the event was handled.
        """
        if (
            event.type() == QtCore.QEvent.Type.PaletteChange
            or event.type() == QtCore.QEvent.Type.ApplicationPaletteChange
        ):
            self.on_recolour()
        return super().event(event)


class nexafsConverterQANT(nexafsParserConverter):
    def __init__(self, parsers: list[Type[parserBase]] = [], parent=None):
        super().__init__(parsers, parent)
        self._title_label.setText("QANT Conversion:")

    pass


if __name__ == "__main__":
    app = QtWidgets.QApplication(sys.argv)
    main = nexafsParserConverter()
    main.setWindowTitle("Converter")
    main.show()
    app.exec()



"""
Parser classes for the Medium Energy X-ray 2 (MEX1) beamline at the Australian Synchrotron.
"""

# Internal
from pyNexafs.parsers import parserBase, parserMeta
from pyNexafs.utils.mda import MDAFileReader
from pyNexafs.utils.reduction import reducer
from pyNexafs.parsers.au.aus_sync.MEX1_relabels import RELABELS

# Standard
from io import TextIOWrapper
from typing import Any
import ast
import warnings
import datetime as dt
import os

# External
from numpy.typing import NDArray
import numpy as np

has_PYQT: bool
try:
    from PyQt6 import QtWidgets

    has_PYQT = True
    from pyNexafs.gui.widgets.reducer import EnergyBinReducerDialog
except ImportError:
    has_PYQT = False

# TODO: What are the values for MEX1? What are the bins?
# Additional data provided by the MEX1 beamline for the data reduction
# BIN_ENERGY_DELTA = 11.935
# BIN_96_ENERGY = 1146.7
TOTAL_BINS = 4096
# TOTAL_BIN_ENERGIES = np.linspace(
#     start=BIN_96_ENERGY - 95 * BIN_ENERGY_DELTA,
#     stop=BIN_96_ENERGY + (TOTAL_BINS - 96) * BIN_ENERGY_DELTA,
#     num=TOTAL_BINS,
# )
INTERESTING_BINS_IDX = [80, 2000]  # 80 to 900 for MEX2.
# INTERESTING_BINS_ENERGIES = TOTAL_BIN_ENERGIES[
# INTERESTING_BINS_IDX[0] : INTERESTING_BINS_IDX[1]
# ]


class MEX1_NEXAFS_META(parserMeta):
    def __init__(
        cls: type,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwds: Any,
    ) -> "MEX1_NEXAFS":
        # Create an extra class property for MEX1 mda data, to track binning settings
        cls.reduction_bin_domain: list[tuple[int, int]] | None = None
        """Tracker for the binning settings used in the most recent data reduction."""

        # Perform the normal class creation via parserMeta.
        return super().__init__(name=name, bases=bases, namespace=namespace, **kwds)


class MEX1_NEXAFS(parserBase, metaclass=MEX1_NEXAFS_META):
    """
    Australian Synchrotron Soft X-ray (SXR) NEXAFS parser.

    WARNING: This parser is experimental, and may not be fully functional.

    Parses data formats including '.asc' and '.mda' formats from the SXR
    Near Edge X-ray Absorption Fine Structure (NEXAFS) tool.

    Parameters
    ----------
    filepath : str | None
        The file path to the data file.
    header_only : bool, optional
        If True, only the header of the file is read, by default False.
    relabel : bool | None, optional
        If True, then the parser will relabel the data columns, by default None.
    energy_bin_range : tuple[float, float] | None = None
        The energy range to bin the data, by default None.
    use_recent_binning : bool, optional
        If True, then the '.mda' parsers will use the most recent class
        reduction binning settings. By default True, as UIs will use this,
        and assume the parser will use the same binning settings.
        TODO Perhaps make this a property of the base class, that way UIs can
        reset such settings if needed upon directory change etc.

    Attributes
    ----------
    ALLOWED_EXTENSIONS
    SUMMARY_PARAM_RAW_NAMES
    COLUMN_ASSIGNMENTS
    RELABELS

    Notes
    -----
    Implemented for data as of 2024-Mar.
    """

    ALLOWED_EXTENSIONS = [".xdi", ".mda"]
    SUMMARY_PARAM_RAW_NAMES = [
        "Sample|Comment 1",
        "AngleElement.symbol",
        "Element.edge",
        "E1",
        "E2",
        "E3",
        "E4",
    ]
    COLUMN_ASSIGNMENTS = {
        "x": "Energy|Energy Setpoint|energy",
        "y": [
            "Bragg|bragg",
            "Fluorescence|iflour|Fluorescence Sum",
            "Count Time|count_time",
            "I0|i0",
            "Sample Drain|SampleDrain",
            "ICR_AVG",
            "OCR_AVG",
        ],
        "y_errs": None,
        "x_errs": None,
    }

    RELABELS = RELABELS

    def __init__(
        self,
        filepath: str | None,
        header_only: bool = False,
        relabel: bool | None = None,
        energy_bin_range: tuple[float, float] | None = None,
        use_recent_binning: bool = True,
    ) -> None:
        # Manually add kwargs
        kwargs = {}
        if use_recent_binning is not None:
            kwargs.update(use_recent_binning=use_recent_binning)
        if energy_bin_range is not None:
            kwargs.update(energy_bin_range=energy_bin_range)
        super().__init__(filepath, header_only, relabel, **kwargs)

    @classmethod
    def parse_xdi_2024_08(
        cls,
        file: TextIOWrapper,
        header_only: bool = False,
    ) -> tuple[NDArray, list[str], list[str], dict[str, Any]]:
        """
        Read Australian Synchrotron MEX1 .xdi files, as of 2024-Aug.

        Parameters
        ----------
        file : TextIOWrapper
            TextIOWrapper of the datafile (i.e. open('file.xdi', 'r')).
        header_only : bool, optional
            If True, then only the header of the file is read and NDArray is returned as None, by default False.

        Returns
        -------
        tuple[NDArray, list[str], dict[str, Any]]
            Returns a set of data as a numpy array,
            labels as a list of strings,
            units as a list of strings,
            and parameters as a dictionary.

        Raises
        ------
        ValueError
            If the file is not a valid .xdi file.
        """
        # Initialise structures
        params = {}
        labels = []
        units = None

        # Check valid format.
        if not file.name.endswith(".xdi"):
            raise ValueError(f"File {file.name} is not a valid .xdi file.")

        ### Read file
        # Check header is correct
        assert file.readline() == "# XDI/1.0\n"

        ## 1 Initial Parameters
        column_descriptions = {}
        column_type_assignments = {}
        # Read first param line.
        line = file.readline()
        i = 0
        while "# " == line[0:2] and line != "# ///\n":
            line = line[2:].strip().split(": ", 1)  # split first colon
            param, value = tuple(line)

            # Categorise information
            if param.startswith("Column."):
                # Label name
                labels.append(value)
                params[param] = value
            elif param.startswith("Describe.Column."):
                # Label description. No dedicated structure for this, add to params.
                col_name = params[param.replace("Describe.", "")]
                if col_name + " - " in value:
                    value = value.replace(col_name + " - ", "")
                if " -- " in value:
                    value, descr = value.split(" -- ", 1)
                else:
                    descr = None
                params[param] = value
                column_descriptions[i] = (
                    value,
                    col_name + " -- " + descr if descr is not None else col_name,
                    "",  # No units
                )
                column_type_assignments[i] = f"1-D Detector{i:4}"
            else:
                if param == "Samples":
                    # Parse the datapoint length list
                    samples = ast.literal_eval(value)
                    samples = [x.strip() if isinstance(x, str) else x for x in samples]
                    # Add to params
                    for i in range(len(samples)):
                        params[f"Datapoints.Column.{i}"] = samples[i]
                else:
                    # General parameter, add to params.
                    params[param] = value
            # Load new param line
            line = file.readline()
            i += 1

        assert line == "# ///\n"  # Check end of initial parameters

        # Get samplename
        line = file.readline()
        sample_name = line[2:].strip()
        params["Sample"] = sample_name

        # Skip header lines before data
        assert file.readline() == "# \n"
        line = file.readline()
        # Some xdi conversion have a "default mda2xdi" line.
        try:
            assert line == "# xdi from default mda2xdi preset for mex1.\n"
            line = file.readline()
        except AssertionError:
            pass
        # Some xdi files have error messages
        while (
            line.strip()
            == "# All values in column i1 replaced with 1 because consecutive negatives were detected."
        ):
            line = file.readline()
        # Always ends with a line of "#--------\n" before data.
        assert line == "#--------\n"

        # Read data columns
        header_line = file.readline()
        assert header_line[0:2] == "# "
        labels = (
            header_line[2:].strip().split()
        )  # split on whitespace, even though formatting seems to use "   " (i.e. three spaces).
        labels = [label.strip() if type(label) is str else label for label in labels]

        if header_only:
            # Do not process remaining lines
            return None, labels, units, params

        # Read data
        lines = file.readlines()  # read remaining lines efficiently

        # Convert data to numpy array.
        data = np.loadtxt(lines)
        data = np.array(data)

        return data, labels, units, params

    @classmethod
    def parse_mda_2024_08(
        cls,
        file: TextIOWrapper,
        header_only: bool = False,
        energy_bin_range: tuple[float, float] | None = None,
        use_recent_binning: bool = False,
    ) -> tuple[NDArray, list[str], list[str], dict[str, Any]]:
        """
        Read an Australian Synchrotron MEX1 .mda files.

        Created for data as of 2024-Aug.

        Parameters
        ----------
        file : TextIOWrapper
            TextIOWrapper of the datafile (i.e. open('file.mda', 'r')).
        header_only : bool, optional
            If True, then only the header of the file is read and
            NDArray is returned as None, by default False.
        energy_bin_range : tuple[float, float] | None, optional
            The energy range to bin the data, by default None.
        use_recent_binning : bool, optional
            If True, then the most recent binning settings are used
            for data reduction, by default False.
            Ignored if `energy_bin_range` is specified.

        Returns
        -------
        tuple[NDArray, list[str], dict[str, Any]]
            Returns a set of data as a numpy array,
            labels as a list of strings,
            units as a list of strings,
            and parameters as a dictionary.

        Raises
        ------
        ValueError
            If the file is not a valid .mda file.
        """
        # Initialise parameter list
        params = {}
        labels = []
        units = []

        # Check valid format.
        if not file.name.endswith(".mda"):
            raise ValueError(f"File {file.name} is not a valid .mda file.")

        # Need to reopen the file in byte mode.
        file.close()
        mda = MDAFileReader(file.name)
        mda_header = mda.read_header_as_dict()
        ## Previously threw error for higher dimensional data, now just a warning.
        mda_params = mda.read_parameters()
        mda_arrays, mda_scans = mda.read_scans(header_only=header_only)

        # Add values to params dict
        params.update(mda_header)
        params.update(mda_params)

        # Initialise to None for header only reading.
        mda_1d = None
        # Add column types and descriptions to params.
        if not header_only:
            # 1D array essential
            mda_1d = mda_arrays[0]
            # 2D array optional
            if len(mda_arrays) > 1:
                mda_2d = mda_arrays[1]
                mda_2d_scan = mda_scans[1]
                # Check 'multi-channel-analyser-spectra of fluorescence-detector' names are as expected
                florescence_labels = [
                    "MEX1ES01DET01:MCA1:ArrayData",
                    "MEX1ES01DET01:MCA2:ArrayData",
                    "MEX1ES01DET01:MCA3:ArrayData",
                    "MEX1ES01DET01:MCA4:ArrayData",
                ]
                assert mda_2d_scan.labels() == florescence_labels

                # Take properties from 1D and 2D arrays:
                energies = mda_1d[:, 0] * 1000  # Convert keV to eV for MEX beamline
                dataset = mda_2d[
                    :, INTERESTING_BINS_IDX[0] : INTERESTING_BINS_IDX[1], :
                ]
                bin_e = None  # INTERESTING_BINS_ENERGIES  # pre-calibrated.

                ## Perform binning on 2D array:
                # Is an existing binning range available?
                if energy_bin_range is not None:
                    # Update the class variable with the new binning settings.
                    cls.reduction_bin_domain = energy_bin_range
                    red = reducer(energies, dataset, bin_e)
                elif use_recent_binning and cls.reduction_bin_domain is not None:
                    # Uses the most recent binning settings.
                    red = reducer(energies, dataset, bin_e)
                else:
                    if has_PYQT:
                        # Create a QT application to run the dialog.
                        if QtWidgets.QApplication.instance() is None:
                            QtWidgets.QApplication([])
                        # Run the Bin Selector dialog
                        window = EnergyBinReducerDialog(
                            energies=energies, dataset=dataset, bin_energies=bin_e
                        )
                        # window.reducerUI.bin_axes.set_xlabel("")
                        window.show()
                        if window.exec():
                            # If successful, store the binning settings and data reducer
                            cls.reduction_bin_domain = window.domain
                            red = window.reducer
                        else:
                            raise ValueError("No binning settings selected.")
                    else:
                        raise ValueError(
                            "No binning settings provided, and no PyQt available to select binning settings."
                        )
                        # TODO: Add a command line interface.

                # Use the binning settings to reduce the data.
                _, reduced_summed_data = red.reduce_by_sum(
                    bin_domain=cls.reduction_bin_domain,
                    axis=None,  # all axes
                )
                _, reduced_single_detector_data = red.reduce_by_sum(
                    bin_domain=cls.reduction_bin_domain,
                    axis="bin_energies",  # just the bin energies
                )

            # 3D array unhandled.
            if len(mda_arrays) > 2:
                warnings.warn(
                    "MDA file(s) contain more than two dimension, handling of higher dimensions is not yet implemented."
                )
        # Collect units and labels:
        scan_1d = mda_scans[0]
        positioners = scan_1d.positioners
        detectors = scan_1d.detectors
        if len(mda_scans) > 1:
            scan_2d = mda_scans[1]
            positioners += scan_2d.positioners
            detectors += scan_2d.detectors

        column_types = {
            "Positioner": (
                "name",
                "descr",
                "step mode",
                "unit",
                "rdbk name",
                "rdbk descr",
                "rdbk unit",
            ),
            "Detector": (
                "name",
                "descr",
                "unit",
            ),
        }
        column_descriptions = {
            i: [
                p.name,
                p.desc,
                p.step_mode,
                p.unit,
                p.readback_name,
                p.readback_desc,
                p.readback_unit,
            ]
            for i, p in enumerate(positioners)
        }
        column_descriptions.update(
            {
                i + len(positioners): [d.name, d.desc, d.unit]
                for i, d in enumerate(detectors)
            }
        )
        params["column_types"] = column_types
        params["column_descriptions"] = column_descriptions

        # Collect units and labels:
        for i, p in enumerate(positioners):
            labels.append(p.name)
            units.append(p.unit)
        for i, d in enumerate(detectors):
            labels.append(d.name)
            units.append(d.unit)
        # If 2D data is present, add reduced data to 1D data.
        if not header_only and len(mda_arrays) > 1:
            # Check rows (energies) match length
            assert reduced_summed_data.shape[0] == mda_1d.shape[0]
            assert reduced_single_detector_data.shape[0] == mda_1d.shape[0]
            # Add reduced data to 1D data as extra columns
            mda_1d = np.c_[mda_1d, reduced_single_detector_data, reduced_summed_data]
            # Add labels and units for reduced data
            # (Detector data labels/units already added via positioners and detectors above.)
            labels += ["MCA Sum (Reduced)"]
            units += ["a.u."]
        # Use scan time if available, otherwise let system time be used.
        if "MEX1ES01GLU01:MEX_TIME" in params:
            params["created"] = params["MEX1ES01GLU01:MEX_TIME"]
        return mda_1d, labels, units, params


def MEX1_to_QANT_AUMainAsc(
    parser: parserBase,
    extrainfo_mapping={
        "SR14ID01MCS02FAM:X.RBV": None,  # "Mono.d_spacing",
        "SR14ID01MCS02FAM:Y.RBV": None,
        "SR14ID01MCS02FAM:Z.RBV": None,
        "SR14ID01MCS02FAM:R1.RBV": None,
        "SR14ID01MCS02FAM:R2.RBV": None,
        "SR14ID01NEXSCAN:saveData_comment1": "Sample",
        "SR14ID01NEXSCAN:saveData_comment2": None,  # "Reference.Element",
    },
) -> list[str]:
    """
    Convert a MEX1 parser mapping to to QANT format.

    Parameters
    ----------
    parser : parserBase
        The parser object (with data, labels, units, and params loaded) to convert.
    extrainfo_mapping : dict[str:str|None], optional
        Optional mapping for known read-values for the QANT AUMainAsc format to
        parser parameter names. By default the dictionary key values are:
            {"SR14ID01MCS02FAM:X.RBV": None,
            "SR14ID01MCS02FAM:Y.RBV": None,
            "SR14ID01MCS02FAM:Z.RBV": None,
            "SR14ID01MCS02FAM:R1.RBV": None,
            "SR14ID01MCS02FAM:R2.RBV": None,
            "SR14ID01NEXSCAN:saveData_comment1": None,
            "SR14ID01NEXSCAN:saveData_comment2": None,}

    Returns
    -------
    list[str]
        A list of lines for the QANT AUMainAsc format, with newline terminations included.
    """
    possible_read_values = [
        "SR14ID01MCS02FAM:X.RBV",
        "SR14ID01MCS02FAM:Y.RBV",
        "SR14ID01MCS02FAM:Z.RBV",
        "SR14ID01MCS02FAM:R1.RBV",
        "SR14ID01MCS02FAM:R2.RBV",
        "SR14ID01NEXSCAN:saveData_comment1",
        "SR14ID01NEXSCAN:saveData_comment2",
    ]
    # Check validity of the extrainfo_mapping
    for key, value in extrainfo_mapping.items():
        if value is not None:
            if value not in parser.params:
                raise ValueError(f"Parameter {value} not found in parser params.")
            elif key not in possible_read_values:
                raise ValueError(f"Parameter {key} not found in possible read values.")
    # Create reverse dict
    extrainfo_remapping = {v: k for k, v in extrainfo_mapping.items()}

    # Check vailidty of parser object:
    if parser.data is None:
        raise ValueError("Parser object does not have data loaded.")

    # Check validity of the dimensionality of the object:
    if len(parser.data.shape) != 2:
        raise ValueError("Parser object data is not 2D.")

    # Define a container for the output strings, line by line.
    ostrs = []

    # Define fake asc version number:
    ostrs.append("## mda2ascii 0.3.2 generated output\n")
    ostrs.append("\n")
    ostrs.append("\n")

    # Rename for consistency between mda and asc formats.
    mda_param_names = ["mda_version", "mda_scan_number", "mda_rank", "mda_dimensions"]
    asc_param_names = [
        "MDA File Version",
        "Scan Number",
        "Overall scan dimension",
        "Total requested scan size",
    ]
    for mda, asc in zip(mda_param_names, asc_param_names):
        if mda in parser.params:
            parser.params[asc] = parser.params[mda]
            del parser.params[mda]

    # Define MDA versioning from parameters, or create fictitious version.
    ostrs.append(
        "# MDA File Version = 1.3\n"
        if "MDA File Version" not in parser.params
        else f"# MDA File Version = {parser.params['MDA File Version']}\n"
    )
    ostrs.append(
        "# Scan Number = 1\n"
        if "Scan Number" not in parser.params
        else f"# Scan Number = {parser.params['Scan Number']}\n"
    )
    ostrs.append(
        "# Overall scan dimension = 1-D\n"
        if "Overall scan dimension" not in parser.params
        else f"# Overall scan dimension = {parser.params['Overall scan dimension']}-D\n"
    )
    ostrs.append(
        f"# Total requested scan size = {len(parser.data)}\n"
        if "Total requested scan size" not in parser.params
        else f"# Total requested scan size = {parser.params['Total requested scan size']}\n"
    )
    ostrs.append("\n")
    ostrs.append("\n")

    # Define the extra PVs
    ostrs.append("#  Extra PV: name, descr, values (, unit)\n")
    ostrs.append("\n")
    param_idx = 1
    for param in parser.params:
        if param not in asc_param_names:
            wparam = (
                param
                if param not in extrainfo_remapping
                or extrainfo_remapping[param] is None
                else extrainfo_remapping[param]
            )
            line = f"# Extra PV {param_idx}: {wparam}"
            if (
                hasattr(parser.params[param], "__len__")
                and type(parser.params[param]) is not str
            ):
                for val in parser.params[param]:
                    wval = (
                        val
                        if val not in extrainfo_remapping
                        or extrainfo_remapping[val] is None
                        else extrainfo_remapping[val]
                    )
                    line += f", {wval}"
                line += "\n"
            else:
                line += f", {parser.params[param]}, \n"
            ostrs.append(line)
            param_idx += 1
    ostrs.append("\n")
    ostrs.append("\n")

    # Define the scan header:
    ostrs.append("# 1-D Scan\n")
    ostrs.append(
        f"# Points completed = {parser.data.shape[0]} of {parser.data.shape[0]}\n"
        if "Points completed" not in parser.params
        else f"# Points completed = {parser.params['Points completed']}\n"
    )
    ostrs.append(
        "# Scanner = SR14ID01NEXSCAN:scan1\n"
        if "Scanner" not in parser.params
        else f"# Scanner = {parser.params['Scanner']}\n"
    )
    if "Scan time" in parser.params:
        ostrs.append(f"# Scan time = {parser.params['Scan time']}\n")
    elif "created" in parser.params and isinstance(
        parser.params["created"], dt.datetime
    ):
        created_timestamp: dt.datetime = parser.params["created"]
        ostrs.append(
            f"# Scan time = {created_timestamp.strftime(r'%b %d, %Y %H:%M:%S.%f')}\n"
        )
    elif "modified" in parser.params and isinstance(
        parser.params["modified"], dt.datetime
    ):
        mod_timestamp: dt.datetime = parser.params["modified"]
        ostrs.append(
            f"# Scan time = {mod_timestamp.strftime(r'%b %d, %Y %H:%M:%S.%f')}\n"
        )
    else:
        # Use current time
        ostrs.append(
            f"# Scan time = {dt.datetime.now().strftime(r'%b %d, %Y %H:%M:%S.%f')}\n"
        )
    ostrs.append("\n")

    if "column_types" in parser.params and isinstance(
        parser.params["column_types"], dict
    ):
        for coltype in parser.params["column_types"].keys():
            line = f"#  {coltype}:"
            for val in parser.params["column_types"][coltype]:
                line += f" {val}"
                if val != parser.params["column_types"][coltype][-1]:
                    line += ","
                else:
                    line += "\n"
            ostrs.append(line)
    else:
        # Use default
        ostrs.append(
            "#  Positioner: name, descr, step mode, unit, rdbk name, rdbk descr, rdbk unit\n"
        )
        ostrs.append("#  Detector: name, descr, unit\n")
        ostrs.append("\n")

    # Define the Column Descriptions
    ostrs.append("# Column Descriptions:\n")
    if "column_descriptions" in parser.params:
        # Column descriptions have been saved. Use these.
        init_idx = 1  # default, add 1 to column description indexes
        if parser.params["column_descriptions"][0][0] != "Index":
            # Create index column if not present.
            ostrs.append(f"#{1:5}  [     Index      ]\n")
            init_idx = 2  # Because of added extra index.
        for i, col in enumerate(parser.params["column_descriptions"]):
            if "column_type_assignments" in parser.params:
                col_type = parser.params["column_type_assignments"][i]
            else:
                col_type = f"1-D Detector{i + 1:4}"  # Nth detector
            line = f"#{i + init_idx + 1:5}  [" + col_type + "]  "
            for val in col:
                line += f"{val}" if val is not None else ""
                if val != col[-1]:
                    line += ", "
                else:
                    line += "\n"
            ostrs.append(line)
    else:
        # Create column descriptions from units and labels.
        ostrs.append(f"#{1:5}  [     Index      ]\n")
        for i, label in enumerate(parser.labels):
            line = f"#{i + 2:5}  "
            if i == 0:
                # Assume energy labels...
                line += f"[1-D Positioner 1]  {label}, "
                if "units" in parser.params and parser.params["units"][i] is not None:
                    unit = parser.params["units"][i]
                    line += f"Mono setpoint, TABLE, {unit}, {label}, Mono setpoint, {unit}\n"
                else:
                    line += f"Mono setpoint, TABLE, eV, {label}, Mono setpoint, eV\n"
            else:
                # Assume detector labels, no description.
                line += f"[1-D Detector{i:4}]  {label}, "
                if "units" in parser.params and parser.params["units"][i] is not None:
                    unit = parser.params["units"][i]
                    line += f"{label}, , {unit}\n"
                else:  # no unit.
                    line += f"{label}, , \n"
            ostrs.append(line)
    ostrs.append("\n")

    # Define the Scan Values
    ostrs.append("# 1-D Scan Values\n")
    for i, row in enumerate(parser.data):
        line = f"{i + 1}"
        for val in row:
            line += f"\t{val}"
        line = line[:-1] + "\n"
        ostrs.append(line)

    # End of file
    return ostrs


if __name__ == "__main__":
    # Example usage
    path = os.path.dirname(__file__)
    package_path = os.path.normpath(os.path.join(path, "../../../../"))

    # Open an mdi file
    mda_path = os.path.normpath(
        os.path.join(package_path, "tests/test_data/au/MEX1/MEX1_40747.mda")
    )

    # HEADER
    test1 = MEX1_NEXAFS(mda_path, header_only=True)
    # BODY
    test2 = MEX1_NEXAFS(mda_path, header_only=False)
    # Check if previous binning is applied to new data.
    test3 = MEX1_NEXAFS(mda_path, header_only=False)

    import matplotlib.pyplot as plt

    plt.close("all")
    subplts = plt.subplots(1, 1)
    fig: plt.Figure = subplts[0]
    ax: plt.Axes = subplts[1]
    ave_mca_idx = -1
    energy_idx = test2.label_index("ring_current")
    ax.plot(
        test2.data[:, energy_idx],
        test2.data[:, ave_mca_idx],
        label="Test2" + test2.labels[ave_mca_idx],
    )
    ax.set_xlabel(test2.labels[energy_idx])
    # plt.ioff()

    # Also open an xdi file
    xdi_path = os.path.normpath(
        os.path.join(package_path, "tests/test_data/au/MEX1/MEX1_40747_processed.xdi")
    )
    test4 = MEX1_NEXAFS(xdi_path, header_only=False)
    energy_idx2 = test4.label_index("ring_current")
    ax.plot(
        test4.data[:, energy_idx2],
        test4.data[:, ave_mca_idx],
        label="Test4" + test4.labels[ave_mca_idx],
    )
    print(test4.labels[energy_idx2])

    ax.legend()
    plt.ion()
    # plt.show(block=False)
    plt.show(block=True)



"""
Implementations of typical NEXAFS specturm fitting functions.
"""

import numpy as np
import numpy.typing as npt
import scipy.special as spec

# -------------------------------------------------------------------
####################### Peak Functions ##############################
# -------------------------------------------------------------------


def gauss(x: npt.ArrayLike, amplitude: float, sigma: float, x0: float) -> npt.NDArray:
    """
    Gaussian function for fitting peaks in NEXAFS spectra.

    Parameters
    ----------
    x : ArrayLike
        A list of position values to evaluate the Gaussian on.
    amplitude : float
        The amplitude intensity of the Gaussian.
    sigma : float
        The standard deviation parameter of the Gaussian.
    x0 : float
        The positional offset of the Gaussian.

    Returns
    -------
    np.array
        An array of Gaussian values corresponding to `x` values.
    """
    x = np.asarray(x)
    return amplitude * np.exp((x - x0) ** 2 / sigma**2)


# -------------------------------------------------------------------
####################### Baseline Functions ##############################
# -------------------------------------------------------------------


def edge(x: npt.ArrayLike, amplitude: float, x0: float, sigma: float) -> npt.NDArray:
    """
    A step edge for capturing the elemental absorption edge.

    Combines an error function (erf) with positional and amplitude information.

    Parameters
    ----------
    x : ArrayLike
        A list of position values to evaluate the edge on.
    amplitude : float
        The amplitude intensity of the edge.
    x0 : float
        The positional offset of the erf.
    sigma : float
        The width of the erf step.

    Returns
    -------
    np.array
        An array of edge values corresponding to `x` values.
    """
    x = np.asarray(x)
    return amplitude * spec.erf(x - x0, sigma)


def edge_decaying(
    x: npt.ArrayLike, amplitude: float, x0: float, sigma: float, alpha: float
) -> npt.NDArray:
    """
    A step edge with decay for capturing an elemental absorption edge.

    Uses the same base function as `edge`, but adds an decaying exponent for x > x0.

    Parameters
    ----------
    x : ArrayLike
        A list of position values to evaluate the edge on.
    amplitude : float
        The amplitude intensity of the edge.
    x0 : float
        The positional offset of the erf.
    sigma : float
        The width of the erf step.
    alpha : float
        The magnitude of the exponential decay exponent, applied to x > x0.

    Returns
    -------
    np.array
        An array of edge values corresponding to `x` values.
    """
    x = np.asarray(x)
    y_edge = edge(x, amplitude, x0, sigma)
    # Check positive alpha
    assert alpha >= 0, "The exponent `alpha` must not be negative."

    # Find x > x0
    idxs = x > x0  # At x=x0, the amplitude is 1.
    x_decay = x[idxs]

    # Apply the decay
    y_edge[idxs] *= np.exp(-alpha * (x_decay - x0))

    # Return the result
    return y_edge



"""
Utility functions for handling sizes and conversions.
"""


def btyes_to_human_readable(num_bytes: int) -> str:
    """
    Convert bytes to a human-readable format (KB, MB, GB, etc.).

    Parameters
    ----------
    num_bytes : int
        Number of bytes.

    Returns
    -------
    str
        A human-readable string.
    """
    if num_bytes < 0:
        raise ValueError("Number of bytes must be non-negative.")

    if num_bytes == 0:
        return "0 B"

    size_units = ["B", "KB", "MB", "GB", "TB"]
    index = 0

    while num_bytes >= 1024 and index < len(size_units) - 1:
        num_bytes = num_bytes // 1024
        index += 1

    return f"{num_bytes:.2f} {size_units[index]}"


if __name__ == "__main__":
    # Example usage
    print(btyes_to_human_readable(1024))  # Output: "1.00 KB"
    print(btyes_to_human_readable(1048576))  # Output: "1.00 MB"
    print(btyes_to_human_readable(1073741824))  # Output: "1.00 GB"
    print(btyes_to_human_readable(0))  # Output: "0 B"
    print(btyes_to_human_readable(-1))  # Raises ValueError



"""
Parser classes for the Australian Synchrotron.

Current beamlines supported include are:
- SXR_NEXAFS : Australian Synchrotron Soft X-ray NEXAFS parser.
- MEX2_NEXAFS : Australian Synchrotron Medium Energy X-ray NEXAFS parser.
"""

from pyNexafs.parsers.au.aus_sync.SXR import SXR_NEXAFS
from pyNexafs.parsers.au.aus_sync.MEX1 import MEX1_NEXAFS, MEX1_to_QANT_AUMainAsc
from pyNexafs.parsers.au.aus_sync.MEX2 import MEX2_NEXAFS, MEX2_to_QANT_AUMainAsc

all = [
    SXR_NEXAFS,
    MEX1_NEXAFS,
    MEX1_to_QANT_AUMainAsc,
    MEX2_NEXAFS,
    MEX2_to_QANT_AUMainAsc,
]



"""
This file contains the paths to file resources, used in the GUI or by the package.
"""

import os

root = os.path.dirname(os.path.abspath(__file__))

ICONS = {
    # "normalisation": os.path.normpath(os.path.join(root, "gui/icons/flaticon/normalization_b&w.png")),
    # "normalisation": os.path.normpath(os.path.join(root, "gui/icons/normalisation.png")),
    "normalisation_light": os.path.normpath(
        os.path.join(root, "gui/icons/normalisation2_light.png")
    ),
    "normalisation_dark": os.path.normpath(
        os.path.join(root, "gui/icons/normalisation2_dark.png")
    ),
}



"""
Relabels for the MEX1 data.
"""

RELABELS = {
    # -------------------MDA File-------------------
    # 'MEX1SSCAN01:saveData_comment1':,
    # 'MEX1SSCAN01:saveData_comment2':,
    # 'MEX1SSCAN01:saveData_realTime1D':,
    # 'MEX1SSCAN01:saveData_fileSystem':,
    # 'MEX1SSCAN01:saveData_subDir':,
    # 'MEX1SSCAN01:saveData_fileName':,
    # 'MEX1SSCAN01:scan1.P1SM':,
    # 'MEX1SSCAN01:scan1.P2SM':,
    # 'MEX1SSCAN01:scan1.P3SM':,
    # 'MEX1SSCAN01:scan1.P4SM':,
    # 'MEX1SSCAN01:scanTypeSpec':,
    # 'MEX1SSCAN01:scan1.BSPV':,
    # 'MEX1SSCAN01:scan1.BSCD':,
    # 'MEX1SSCAN01:scan1.BSWAIT':,
    # 'MEX1SSCAN01:scan1.ASPV':,
    # 'MEX1SSCAN01:scan1.ASCD':,
    # 'MEX1SSCAN01:scan1.ASWAIT':,
    # 'MEX1SSCAN01:scan1.PDLY':,
    # 'MEX1SSCAN01:scan1.DDLY':,
    # 'TS01:SECONDS_MONITOR':,
    # 'MEX1ES01GLU01:MEX_TIME':,
    # 'MEX1SLT01:VSIZE.RBV':,
    # 'MEX1SLT01:VCENTRE.RBV':,
    # 'MEX1SLT01:HSIZE.RBV':,
    # 'MEX1SLT01:HCENTRE.RBV':,
    # 'MEX1SLT01:VSIZE.OFF':,
    # 'MEX1SLT01:VCENTRE.OFF':,
    # 'MEX1SLT01:HSIZE.OFF':,
    # 'MEX1SLT01:HCENTRE.OFF':,
    # 'MEX1SLT01MOT01.RBV':,
    # 'MEX1SLT01MOT03.RBV':,
    # 'MEX1SLT01MOT02.RBV':,
    # 'MEX1SLT01MOT04.RBV':,
    # 'MEX1SCRN01LGHT01:BRIGHTNESS_MONITOR':,
    # 'MEX1SCRN01ACTP01:INSERT_WITHDRAW_STATUS':,
    # 'MEX1MIR01:POSITIONER:select.RVAL':,
    # 'MEX1MIR01:TRANS.RBV':,
    # 'MEX1MIR01:YAW.RBV':,
    # 'MEX1MIR01MOT06.RBV':,
    # 'MEX1MIR01:HEIGHT.RBV':,
    # 'MEX1MIR01:ROLL.RBV':,
    # 'MEX1MIR01:PITCH.RBV':,
    # 'MEX1MIR01MOT01.RBV':,
    # 'MEX1MIR01MOT02.RBV':,
    # 'MEX1MIR01MOT03.RBV':,
    # 'MEX1MIR01MOT04.RBV':,
    # 'MEX1MIR01MOT05.RBV':,
    # 'MEX1DCM01:ENERGY_RBV':,
    # 'MEX1DCM01:ENERGY_EV_RBV':,
    # 'MEX1DCM01:OFFSET_RBV':,
    # 'MEX1DCM01:XTAL_INBEAM.RVAL':,
    # 'MEX1DCM01:FINE_PITCH_MRAD_RBV':,
    # 'MEX1DCM01:FINE_ROLL_MRAD_RBV':,
    # 'MEX1DCM01MOT01.RBV':,
    # 'MEX1DCM01MOT02.RBV':,
    # 'MEX1DCM01MOT05.RBV':,
    # 'MEX1DCM01MOT03.RBV':,
    # 'MEX1DCM01MOT04.RBV':,
    # 'MEX1DCM01MOT01.OFF':,
    # 'MEX1DCM01MOT02.OFF':,
    # 'MEX1DCM01:y2_track':,
    # 'MEX1DCM01:y2_mvmin':,
    # 'MEX1DCM01:th_mvmin':,
    # 'MEX1DCM01:Dspace':,
    # 'MEX1DCM01:Mono1110DSpace':,
    # 'MEX1DCM01:Mono1110ThetaOffset':,
    # 'MEX1DCM01:Mono1110HeightOffset':,
    # 'MEX1DCM01:Mono1110Pitch':,
    # 'MEX1DCM01:Mono1110Roll':,
    # 'MEX1DCM01:Mono1110Centre':,
    # 'MEX1DCM01:Mono11130DSpace':,
    # 'MEX1DCM01:Mono11130ThetaOffset':,
    # 'MEX1DCM01:Mono11130HeightOffset':,
    # 'MEX1DCM01:Mono11130Pitch':,
    # 'MEX1DCM01:Mono11130Roll':,
    # 'MEX1DCM01:Mono11130Centre':,
    "MEX1ES01DET01:MCA1:ArrayData": "MCA Ch1",
    "MEX1ES01DET01:MCA2:ArrayData": "MCA Ch2",
    "MEX1ES01DET01:MCA3:ArrayData": "MCA Ch3",
    "MEX1ES01DET01:MCA4:ArrayData": "MCA Ch4",
    # -------------------XDI File-------------------
    # 'Facility.name':,
    # 'Beamline.name':,
    # 'Mono.d_spacing':,
    # 'Element.symbol':,
    # 'Element.edge':,
    # 'Processing.version':,
    # 'Reference.Element':,
    # 'ROI.start_bin':,
    # 'ROI.end_bin':,
    # 'File.Name':,
    "MEX1ES01ZEB01:CALC_ENERGY_EV": "energy",
    "MEX1ES01ZEB01:BRAGG_WITH_OFFSET": "bragg",
    "SR11BCM01:CURRENT_MONITOR": "ring_current",
    "MEX1ES01ZEB01:GATE_TIME_SET": "count_time",
    "MEX1ES01DAQ01:ch1:S:MeanValue_RBV": "i0",
    "MEX1ES01DAQ01:ch2:S:MeanValue_RBV": "i1",
    "MEX1ES01DAQ01:ch3:S:MeanValue_RBV": "i2",
    "MEX1ES01DET01:C1SCA:0:Value_RBV": "S1_clock_ticks",
    "MEX1ES01DET01:C2SCA:0:Value_RBV": "S2_clock_ticks",
    "MEX1ES01DET01:C3SCA:0:Value_RBV": "S3_clock_ticks",
    "MEX1ES01DET01:C4SCA:0:Value_RBV": "S4_clock_ticks",
    "MEX1ES01DET01:C1SCA:5:Value_RBV": "S1_window1",
    "MEX1ES01DET01:C2SCA:5:Value_RBV": "S2_window1",
    "MEX1ES01DET01:C3SCA:5:Value_RBV": "S3_window1",
    "MEX1ES01DET01:C4SCA:5:Value_RBV": "S4_window1",
    "MEX1ES01DET01:C1SCA:8:Value_RBV": "S1_DTFactor",
    "MEX1ES01DET01:C2SCA:8:Value_RBV": "S2_DTFactor",
    "MEX1ES01DET01:C3SCA:8:Value_RBV": "S3_DTFactor",
    "MEX1ES01DET01:C4SCA:8:Value_RBV": "S4_DTFactor",
    "S1_real_time - S1_clock_ticks / 80MHz (clock rate xspress3 mini)": "S1_real_time (clock_ticks / 80MHz)",
    "S2_real_time - S2_clock_ticks / 80MHz (clock rate xspress3 mini)": "S2_real_time",
    "S3_real_time - S3_clock_ticks / 80MHz (clock rate xspress3 mini)": "S3_real_time",
    "S4_real_time - S4_clock_ticks / 80MHz (clock rate xspress3 mini)": "S4_real_time",
    "S1_ifluor - S1_window1 * S1_DTFactor / S1_real_time": "S1_ifluor (window * dead_time / real_time)",
    "S2_ifluor - S2_window1 * S2_DTFactor / S2_real_time": "S2_ifluor",
    "S3_ifluor - S3_window1 * S3_DTFactor / S3_real_time": "S3_ifluor",
    "S4_ifluor - S4_window1 * S4_DTFactor / S4_real_time": "S4_ifluor",
    "avg_ifluor - (ifluor_S1 + ifluor_S2 + ifluor_S3 + ifluor_S4)/4 -- deadtime correction and time normalisation per sensor": "ifluor_avg_corr",
    "S1_mufluor - S1_ifluor / i0": "S1_mufluor (normalised by i0)",
    "S2_mufluor - S2_ifluor / i0": "S2_mufluor",
    "S3_mufluor - S3_ifluor / i0": "S3_mufluor",
    "S4_mufluor - S4_ifluor / i0": "S4_mufluor",
    "avg_mufluor - ifluor_avg_corr / i0 -- deadtime correction and time normalisation per sensor, normalised by i0": "mufluor_avg_corr",
}



import sys

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QWidget,
    QSplitter,
)


from pyNexafs.gui.widgets.fileloader import nexafsFileLoader
from pyNexafs.gui.widgets.viewer import nexafsViewer


class browserWidget(QWidget):
    """A widget for browsing and loading NEXAFS files."""

    def __init__(self, parent=None, init_dir: str | None = None):
        super().__init__(parent=parent)
        # Initialise elements
        self._draggable = QSplitter(Qt.Orientation.Horizontal)
        self._main_layout = QHBoxLayout()
        self.loader = nexafsFileLoader(parent=self)
        self.viewer = nexafsViewer(parent=self)
        # self.converter = converterWidget()

        # Add to layout
        self._draggable.addWidget(self.loader)
        self._draggable.addWidget(self.viewer)
        self._main_layout.addWidget(self._draggable)
        # self.main_layout.addWidget(self.loader)
        # self.main_layout.addWidget(self.viewer)
        # self.sub_layout.addWidget(self.converter)
        self.setLayout(self._main_layout)
        if parent is not None:
            self.setContentsMargins(0, 0, 0, 0)
            self._main_layout.setContentsMargins(0, 0, 0, 0)

        ### Init UI
        # # TODO: Find another way to define a black line on handles with some margin two widgets.
        # grad = QLinearGradient(0, 0, 1, 0)
        # grad.setCoordinateMode(grad.CoordinateMode.ObjectMode)
        # for handle in self.draggable.findChildren(QSplitterHandle):
        #     if handle.parent().orientation() == Qt.Orientation.Horizontal:
        #         # handle.setStyleSheet("background-color: black;")
        #         # handle.setStyleSheet("qlineargradient(spread:pad, x1:0, y1:0, x2:1, y2:0, stop:0 rgba(255, 255, 255, 0), stop:0.407273 rgba(200, 200, 200, 255), stop:0.4825 rgba(101, 104, 113, 235), stop:0.6 rgba(255, 255, 255, 0));")
        #         # handle.setStyleSheet("qlineargradient(x1:0, y1:0, x2:1, y2:1, stop:0 #eee, stop:1 #ccc);")

        #         handle.set

        #         handle.setMinimumWidth(1)
        #         handle.setMinimumHeight(1)
        #         handle.setFixedWidth(1)

        ### Connections
        # Load scans
        self.loader.selectionLoaded.connect(self._on_load_selection)
        # Remove scans with parser / directory change.
        self.loader.directory_selector.new_path.connect(self._on_dir_parser_change)
        self.loader.nexafs_parser_selector.currentIndexChanged.connect(
            self._on_dir_parser_change
        )
        self.loader.relabelling.connect(self.viewer.on_relabel)

        ### Init directory
        if init_dir is not None:
            self.loader.directory_selector.folder_path_edit.setText(init_dir)
            self.loader.directory_selector.folder_path_edit.editingFinished.emit()

    def _on_dir_parser_change(self):
        # Delete the current scan objects
        del self.viewer.scans

    def _on_load_selection(self):
        # Load in the scan objects
        selection_parse_objs = self.loader.loaded_parser_files_selection
        self.viewer.add_parsers_to_scans(selection_parse_objs)
        # Change the file selection in the viewer.
        self.viewer.selected_filenames = self.loader.selected_filenames


def gui(directory: str | None = None):
    app = QApplication(sys.argv)
    window = browserWidget(init_dir=directory)
    window.show()
    window.setWindowTitle("pyNexafs File Browser")
    app.exec()


if __name__ == "__main__":
    gui()



"""
Parser classes for NEXAFS data, including support for synchrotron specific beamlines.

Parsers are registered in the parser_loaders dictionary, which is used to register parsers
in GUI applications.
"""

# Base objects
from pyNexafs.parsers._base import parserBase, parserMeta

# Specific Parsers
from pyNexafs.parsers import au

# Define loaders with string representation.
parser_loaders = {
    "au SXR:NEXAFS": au.SXR_NEXAFS,
    "au MEX1:NEXAFS": au.MEX1_NEXAFS,
    "au MEX2:NEXAFS": au.MEX2_NEXAFS,
}

# Check that all parsers are subclasses of the base parser.
for parser in parser_loaders.values():
    assert issubclass(parser, parserBase), f"{parser} is not a subclass of parserBase."

# Check that no parser names overlap. Important for GUIs.
parser_names = [parser_name for parser_name in parser_loaders.keys()]
assert len(parser_names) == len(set(parser_names)), "Parser names overlap."

__all__ = [
    "parserBase",
    "parserMeta",
    "parser_loaders",
    "au",
]



"""
Definitions of NEXAFS data types, to allow attribute access on nexafs `scan` objects.
"""

from enum import StrEnum
from typing import Any, TypedDict
from typing_extensions import NotRequired
import numpy.typing as npt


class dtype(StrEnum):
    """
    An enumerate for the different NEXAFS data sources.
    """

    ## The direct absorption of X-ray intensity:
    T = "Transmission"  # Equivalent to absorption.
    """Transmission, where the intensity of X-rays transmitted through the sample is measured."""
    A = "Absorption"  # Equivalent to transmission.
    """Absorption, where the absorption of X-rays by the sample is measured."""
    ## Partial photo-electron yield
    PEY = "Partial Electron Yield"
    """Partial Electron Yield, where electrons are collected within a partial energy/momentum window."""
    # The current from a sample drain:
    TEY = "Total Electron Yield"
    """Total Electron Yield, where the total current from the sample drain is measured."""
    # Auger electron yield:
    AEY = "Auger Electron Yield"
    """Auger Electron Yield, where Auger electrons are collected."""

    # Fluorescence yield:
    TFY = "Total Fluorescence Yield"
    """Total Fluorescence Yield, where the total fluorescence from the sample is measured."""

    # Fluorescence yield from a partial energy window:
    PFY = "Partial Fluorescence Yield"
    """Partial Fluorescence Yield, where the fluorescence from a partial energy window is measured."""

    # Normalisation channels:
    I0 = "I0"  # Incident X-ray intensity
    """I0, where the incident X-ray intensity is measured typically via a mesh current, often used for normalisation."""

    PhD = "Photodiode"  # Photodiode scan without sample
    """Photodiode, where the X-ray intensity is measured by a photodiode without the sample, often used for normalisation at the carbon edge."""

    # Energy channels:
    E = "Energy"
    """Energy, the x-ray photon energy."""


# Define the column assignments dictionary typing key-value pairs
class assignments_type(TypedDict):
    """
    Required assignments types for the COLUMN_ASSIGNMENTS parser class property.

    Allows for the key assignments of
    - 'x': A single string or tuple of synonymous strings, identifying the independent variable.
    - 'y': A single string, a tuple of strings, or a list of the former two, identifying dependent variables.
    - 'x_errs': A single string or None, identifying the independent variable error.
    - 'y_errs': A single string, a tuple of strings, or a list of the former two, identifying dependent variable errors.
                If provided, must match the structure of 'y'.
    """

    x: str | tuple[str, ...]
    y: str | tuple[str, ...] | list[str | tuple[str, ...]]
    x_errs: NotRequired[str | tuple[str, ...] | None]
    y_errs: NotRequired[
        str | tuple[str, ...] | None | list[str | tuple[str, ...] | None]
    ]


parse_fn_ret_type = tuple[
    npt.NDArray | tuple[npt.NDArray, ...] | None,  # Data
    list[str | None]
    | list[str]
    | tuple[list[str | None] | list[str], ...]
    | None,  # Labels
    list[str | None]
    | list[str]
    | tuple[list[str | None] | list[str], ...]
    | None,  # Units
    dict[str, Any],  # Params
]
"""
The return type of a parser function.

A tuple that is comprised of:
- data: npt.NDArray | tuple[npt.NDArray, ...] | None
    The data array, or a tuple of data arrays, or None if no data is returned.
- labels: list[str | None] | list[str] | tuple[list[str | None] | list[str], ...] | None
    The labels of the data array(s), or None if no data is returned.
- units: list[str | None] | list[str] | tuple[list[str | None] | list[str], ...] | None
    The units of the data array(s), or None if no data is returned.
- params: dict[str, Any]
    A dictionary of parameters extracted from the file, or an empty dictionary if no parameters are returned.
"""

reduction_type = tuple[
    npt.NDArray | None, list[str | None] | None, list[str | None] | None
]
"""
The return type of a parser reduction function.

A tuple that is comprised of:
- data: npt.NDArray | None
    The reduced data array in a 1D (singular)/2D (multiple) format, or None if no data is returned.
    First index must always correspond to energy, and the second index is the channel indexes (if multiple).
- labels: list[str | None] | None
    The labels of the reduced data array, or None if no data is returned.
- units: list[str | None] | None
    The units of the reduced data array, or None if no data is returned.
"""

if __name__ == "__main__":
    print("Done")
